/* SofChat (Web Chat program SHM version)      */
/* Copyright(c) 1996-2000 Y.YONEYAMA           */
/*                                             */
/* chat.c                                      */

/*---------------------------------------------*/
int version    = 133;    /* current version    */
int patchlevel =   4;    /* patch level        */
int memversion = 132;    /* memory compatible version */
char *software_name = "SofChat";
/*---------------------------------------------*/

#include "config.h"

/*---------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#ifdef WIN32
#include <io.h>
#include "win32/win32.h"
#include "win32/shm.h"
#else /* !WIN32 */
#include <sys/ipc.h>
#include <sys/shm.h>
#ifdef USE_SEM
#include <sys/sem.h>
#endif /* USE_SEM */
#include <unistd.h>
#endif /* WIN32 */

#include "sofchat.h"
#include "utils.h"
#include "html.h"
#include "messages.h"
#include "code.h"
#include "export.h"
#ifndef NOROMAKANA
#include "japanese.h"
#endif /* NOROMAKANA */


#define MAX_USERS      200
#define MAX_LINES     1000
#define MAX_BBS        500
#define MAX_DENYHOSTS   10
#define MAX_LINKS       10
#define NEWUSER_LINES   15
#define TIMEOUT         30       /* 30 sec. */
#define BLINKTIME       90
#define BOLDTIME        40
#define AUTOLOGOUT     120
#define LOGOUTWAIT       3
#define SLEEPTIME      600       /* (Zzz) 10 min. */
#define REFUSETIME       1       /* default refuse time (1 hour)*/

#define TITLE_LEN      256
#define DEF_MAX_LINES   30
#define DEF_BBSLINES    20
#define DEF_IMODE_MAXLINES  15 
#define DEF_IMODE_BBSLINES  5 

/* chat lock mode */
#define CHAT_LOCK_PUBLIC     0 /* public */
#define CHAT_LOCK_REGISTED   1 /* registed user only */
#define CHAT_LOCK_PERMIT     2 /* permitted user only */
#define CHAT_LOCK_CURRENT    3 /* current logging user only */

typedef struct {
  int  version;                 /* memory version */
  char memlock;                 /* lock status */
  char title[TITLE_LEN];
  char titleimage[ADDR_LEN];
  char url[ADDR_LEN];
  char bgcolor[COLOR_LEN];
  char bgimage1[ADDR_LEN];
  char bgimage2[ADDR_LEN];
  char bgimage3[ADDR_LEN];
  char txcolor[COLOR_LEN];
  char licolor[COLOR_LEN];
  char vlcolor[COLOR_LEN];
  char alcolor[COLOR_LEN];
  char dncolor[COLOR_LEN];      /* default name color */
  char dtcolor[COLOR_LEN];      /* date color */ 
  USERINFO user[MAX_USERS];
  TEXT topic;
  TEXT text[MAX_LINES];         /* text */
  int  text_line;               /* current line */
  TEXT bbs[MAX_BBS];            /* bbs */
  int  bbs_line;                /* current bbs line */
  HOSTINFO denyhosts[MAX_DENYHOSTS];
  LINK links[MAX_LINKS];
} CHAT;

CHAT *get_chat();
void release_chat();
void lock_chat();
void unlock_chat();
void reset_chat();

#ifdef EXPORT_USER
void export_users();
void import_users();
#endif
#ifdef EXPORT_BBS
void export_bbss();
void import_bbss();
#endif
#ifdef EXPORT_INI
void export_ini();
void import_ini();
#endif
#ifdef EXPORT_CHAT
void export_chat();
void import_chat();
#endif

void entry_form();
void password_form();
void enter_base();
void input_form();

void printform();
void printlinks();
void printdebug();
void printchat();
void printtopic();
void printbbs();
void printdate();
void sigalm();
void sigterm();
void killtag();
char *getcgiarg();
void clearuserinfo();
int setuserinfo();
char *getparam();
void paramencode();
TEXT *gettextline();
time_t get_atime();
void show_users();
void header();
void footer();
void scerr();
void event_form();
void print_forward();

#ifndef CHATCGI
#define CHATCGI "chat.cgi"
#endif /* CHATCGI */

const char *cgibin = CHATCGI;
const char *official_url = "http://www.asahi-net.or.jp/~fj9y-ynym/SofChat/";
const char *official_url_imode = "http://sofchat.net/i/";

static char cgimtd[5] = DEFAULT_METHOD;

static char agent[AGENT_LEN]    = "";
static char haddress[HADDR_LEN] = "";
static IP   haddressip          = 0;

static CHAT *chat = NULL, *local = NULL;
static int  outhead = 0;
static int  autoreload = 0;
static int  newtext = 0;
static time_t lastatime = 0;
static time_t lastbtime = 0;
static char *title;
static char def_title[TITLE_LEN];
static char def_bgcolor[COLOR_LEN];
static char def_txcolor[COLOR_LEN];
static char def_bgimage[ADDR_LEN];
static char def_licolor[COLOR_LEN];
static char def_vlcolor[COLOR_LEN];
static char def_alcolor[COLOR_LEN];
static char def_dncolor[COLOR_LEN];
static char def_dtcolor[COLOR_LEN];

/* cgi argument */
static int  refresh                = 0;
static int  maxlines               = 0;
static int  bbsline                = 0;
static char mytitle[TITLE_LEN]     = "";
static char bgcolor[COLOR_LEN]     = "";
static char txcolor[COLOR_LEN]     = "";
static char name[NAME_LEN]         = "";
static char namecolor[COLOR_LEN]   = "";
static char textcolor[COLOR_LEN]   = "";
static char password[PASS_LEN]     = "";
static char address[ADDR_LEN]      = "";
static char referer[ADDR_LEN]      = "";
static char mode[10]               = "";  /* system mode */
static char message[MSG_SIZE]      = "";  /* message */
static char lang[10]               = "";  /* output language */
static char jcode[10]              = "";  /* output japanese code */
static char secure[10]             = "";
static int  room                   = 0;
static int  autoclear              = 0;
static int  noframe                = 0;
static int  ostype                 = OS_UNKNOWN; /* unknown os type */
static int  imode                  = 0;
static int  itag                   = 0;
static int  idate                  = 0;
static int  ilog                   = 0;
static int  slog                   = 0;

static char link_url[MAX_LINKS][ADDR_LEN];
static char link_label[MAX_LINKS][LABEL_LEN];
static int  link_items = 0;

static const char *managerid   = MANAGER;
static const char *managerpass = PASSWORD;
static char rawpassword[25]    = "";
static char cpassword[25]      = "";
static char alertmsg[MEMO_SIZE*2] = "";
#ifdef COOKIE
static char cookie[1024]          = "";
static char c_chatname[NAME_LEN]  = "";
static char c_autoclear[2]        = "";
static char c_refresh[4]          = "";
#endif /* COOKIE */

static time_t     now;
static struct tm *nowtm;

static int userid = -1;
static int userproperty = 0;
static int permission = 0;
static int hasalert = 0;
static int alertbtn = 0;
static int skiptitle = 0;
static int refuse = 0;
static int texthr = 0;
static int back = 0;
static int userreseted = 0;
static int deletebtn = 0;

static CHATMODE chat_mode;
static BODY body;
static int oldmozilla = 0;
static int mozilla6 = 0;
static int loginstat = 0;

void chat_exit(ret)
int ret;
{
  if(chat != NULL)
    release_chat(chat);
  exit(ret);
}

void init_memory(chat)
CHAT *chat;
{
  lock_chat(chat);
  memset((char *)chat, 0, sizeof(CHAT));
  chat->version = memversion;

  strncpy(chat->title, MSG(M_TITLE), TITLE_LEN);
  strncpy(chat->bgcolor, DEFAULT_BG, COLOR_LEN);
  strncpy(chat->txcolor, DEFAULT_TX, COLOR_LEN);
  strncpy(chat->licolor, DEFAULT_LI, COLOR_LEN);
  strncpy(chat->vlcolor, DEFAULT_VL, COLOR_LEN);
  strncpy(chat->alcolor, DEFAULT_AL, COLOR_LEN);
  strncpy(chat->dncolor, DEFAULT_DN, COLOR_LEN);
  strncpy(chat->dtcolor, DEFAULT_DT, COLOR_LEN);
  strncpy(chat->text[0].name, MSG(M_TITLE), NAME_LEN);
  strncpy(chat->text[0].message, MSG(M_INITIALIZED), MSG_SIZE);
  chat->text[0].atime = now;
  chat->text_line ++;

#ifdef EXPORT_INI
  import_ini(chat);
#endif
#ifdef EXPORT_USER
  import_users(chat);
#endif
#ifdef EXPORT_BBS
  import_bbss(chat);
#endif
#ifdef EXPORT_CHAT
  import_chat(chat);
#endif
  unlock_chat(chat);
}

int isdenyhost(chat, ip)
CHAT *chat;
IP ip;
{
  int i;
  for(i=0;i<MAX_DENYHOSTS;i++)
  {
    if((chat->denyhosts[i].atime >= now) &&
       (compareIP(chat->denyhosts[i].address, ip, 0xffffff00) == 0))
      return(1);
  }
  return(0);
}

int ismanager()
{
  if((strcmp(name, managerid) == 0) &&
     (strcmp(password, managerpass) == 0))
    return(-1);
  else
    return(0);
}

int get_tagparam(src, len, arg1, arg2, arg3)
char *src;
int len;
char *arg1;
char *arg2;
char *arg3;
{
  int i;
  char c;
  int len1 = 0, len2 = 0, len3 = 0;
  *arg1 = '\0';
  *arg2 = '\0';
  *arg3 = '\0';

  if(*src != '<' && *src != '/') return(0);
  c = *src;
  src++;
  /* arg1 */
  for(i=0;i<(len-1);i++)
  {
    if(((c == '<') && (*src == '>')) || *src == '\0' || *src == ' ') break;
    *(arg1++) = *(src++);
    len1++;
  }
  *arg1 = '\0';
  if(*src != ' ') return(1);
  while(*src == ' ') src++;

  /* arg2 */
  for(i=0;i<(len-1);i++)
  {
    if(((c == '<') && (*src == '>')) || *src == '\0' || *src == ' ') break;
    *(arg2++) = *(src++);
    len2++;
  }
  *arg2 = '\0';
  if(*src != ' ') return(2);
  while(*src == ' ') src++;

  /* arg3 */
  for(i=0;i<(len-1);i++)
  {
    if(*src == '\0')
    {
      break;
    }
    *(arg3++) = *(src++);
    len3++;
  }
  *arg3 = '\0';
  if((len3 > 0) && (c == '<') && (*(arg3-1) == '>')) *(arg3-1) = '\0';
  return(3);
}

int get_userbypassword(name, address, password)
char *name;
char *address;
char *password;
{
  int i;
  for(i=1;i<MAX_USERS;i++)
  {
    if(strcmp(chat->user[i].password, password) == 0)
    {
      strncpy(name, chat->user[i].name, NAME_LEN);
      strncpy(address, chat->user[i].address, ADDR_LEN);
      return(i);
    }
  }
  return(-1);
}

#define MT_ALERT    0
#define MT_ERROR    1

int sendmessage(buf, maxlen, msg, type)
char *buf;
int maxlen;
char *msg;
int type;
{
  int i, len = 0, pos = 1;
  int msgs;
  msgs = (int)buf[0];
  for(i=0;i<msgs;i++)
  {
    int j, t;
    t = buf[pos++];
    j = strlen(&buf[pos]);
    pos += j + 1;
  }
  len = strlen(msg);
  if(len + pos < maxlen - 2)
  {
    buf[pos++] = (char)type;
    strcpy(&buf[pos], msg);
    buf[len + pos] = '\0';
    buf[0]++;
    return(1);
  }
  return(0);
}

void alertmessages()
{
  int i, j;
  if((j = (int)alertmsg[0]) > 0)
  {
    char *p, type;
    hasalert = 1;
    if(!isIMODE(ostype))
    {
      printf("<SCRIPT LANGUAGE=JavaScript>\n");
      printf("<!--\n");
      printf("function showalert() {\n");
    }
    p = &alertmsg[1];
    for(i=0;i<j;i++)
    {
      char amsg[MEMO_SIZE*2], *q = amsg;
      type = *p++;
      if(type == MT_ALERT) alertbtn = 1;
      while(*p != '\0')
      {
        if(!isIMODE(ostype))
        {
          if(sjis1st(*p))
          {
            *q++ = *p++;
            if(oldmozilla && (*p == '\\'))
            {
              *q++ = '\\';
            }
            *q++ = *p++;
          }
          else
          {
            if(*p == '"') *q++ = '\\';
            *q++ = *p++;
          }
        }
        else
        { /* imode */
          if(sjis1st(*p))
          {
            *q++ = *p++;
            *q++ = *p++;
          }
          else
          {
            if(*p == '\\' && *(p+1) == 'n')
            {
              *q++ = '<';
              *q++ = 'B';
              *q++ = 'R';
              *q++ = '>';
              *(p+1) = ' ';
              p++;
            }
            *q++ = *p++;
          }
        }
      }
      *q = '\0';
      if(!isIMODE(ostype))
      {
        printf("alert(\"%s\");\n", amsg);
      }
      else
      {
        printf("<HR>\n*** %s\n<HR>\n", amsg);
      }
      p++;
    }
    if(!isIMODE(ostype))
    {
      printf("}\n");
      printf("//-->\n");
      printf("</SCRIPT>\n");
    }
  }
}

void historyback()
{
  printf("<SCRIPT LANGUAGE=JavaScript>\n");
  printf("<!--\n");
  printf("history.back()\n");
  printf("//-->\n");
  printf("</SCRIPT>\n");
}

/* mode= 0: undenied 1:denied */
/* return -1: error 0: succeed */
int denyhost(chat, dh, tim, name, mode)
CHAT *chat;
IP dh;
int tim;
char *name;
int mode;
{
  int i;
  for(i=0;i<MAX_DENYHOSTS;i++)
  {
    if(chat->denyhosts[i].address == dh && mode == 0)
    { /* undeny */
      chat->denyhosts[i].atime = 0;
      chat->denyhosts[i].address = 0;
      chat->denyhosts[i].mid = userid;
      return(0);
    }
    if((chat->denyhosts[i].atime < now || chat->denyhosts[i].address == dh) && mode == 1)
    {
      chat->denyhosts[i].atime = now + 3600*tim;
      chat->denyhosts[i].address = dh;
      strncpy(chat->denyhosts[i].name, name, NAME_LEN);
      chat->denyhosts[i].mid = userid;
      return(0);
    }
  }
  return(-1);
}

#ifdef EXPORT_USER
void export_users(chat)
CHAT *chat;
{
  int i;

  /* truncate the export file */
  if(export_open(EXPORT_USER) == -1) return;
  export_close();

  for(i = 1; i < MAX_USERS; i++)
  {
    if(*chat->user[i].password != '\0')
    {
      if(export_user(EXPORT_USER, &chat->user[i], 0) == -1) break;
    }
  }
}
void import_users(chat)
CHAT *chat;
{
  int i = 1;
  if(import_open(EXPORT_USER) != -1)
  {
    while(i < MAX_USERS)
    {
      clearuserinfo(&chat->user[i]);
      if(import_user(&chat->user[i]) == 0) break;
      if(*chat->user[i].name != '\0' &&
         (*chat->user[i].password != '\0' || chat->user[i].permit > 0) &&
         (chat->user[i].count > 3) )
      {
        i++;
      }
    }
    import_close();
  }
}
#endif /* EXPORT_USER */

#ifdef EXPORT_BBS
void export_bbss(chat)
CHAT *chat;
{
  int i;

  if(export_open(EXPORT_BBS) != -1)
  {
    for(i = 0; i < MAX_BBS && i < chat->bbs_line; i++)
    {
      int pline;
      if(chat->bbs_line >= MAX_BBS)
        pline = (chat->bbs_line + i)%MAX_BBS;
      else
        pline = i;
      if(*chat->bbs[pline].message != '\0')
      {
        char msg[1025];
        strncpy(msg, chat->bbs[pline].message, 1024);
        msg[1024] = chat->bbs[pline].flags;
        if(export_text(chat->bbs[pline].name,  msg,
                    chat->bbs[pline].atime) == 0) break;
      }
    }
    export_close();
  }
}

void import_bbss(chat)
CHAT *chat;
{
  int i;
  if(import_open(EXPORT_BBS) != -1)
  {
    chat->bbs_line = 0;
    for(i = 0; i < MAX_BBS; i++)
    {
      char msg[1025];
      if(import_text(chat->bbs[i].name,  msg,
           &chat->bbs[i].atime) > 0)
      {
        strncpy(chat->bbs[i].message, msg, 1024);
        chat->bbs[i].flags = msg[1024];
        chat->bbs_line ++;
      }
      else
        break;
    }
    import_close();
  }
}
#endif /* EXPORT_BBS */

#ifdef EXPORT_INI
void export_ini(chat)
CHAT *chat;
{
  FILE *fp;
  int i;
  fp = fopen(EXPORT_INI, "w");
  if(fp == NULL) return;
  fprintf(fp, "TITLE:%s\n", chat->title);
  fprintf(fp, "TITLEIMAGE:%s\n", chat->titleimage);
  fprintf(fp, "BGCOLOR:%s\n", chat->bgcolor);
  fprintf(fp, "BGIMAGE1:%s\n", chat->bgimage1);
  fprintf(fp, "BGIMAGE2:%s\n", chat->bgimage2);
  fprintf(fp, "BGIMAGE3:%s\n", chat->bgimage3);
  fprintf(fp, "TXCOLOR:%s\n", chat->txcolor);
  fprintf(fp, "LICOLOR:%s\n", chat->licolor);
  fprintf(fp, "VLCOLOR:%s\n", chat->vlcolor);
  fprintf(fp, "ALCOLOR:%s\n", chat->alcolor);
  fprintf(fp, "DNCOLOR:%s\n", chat->dncolor);
  fprintf(fp, "DTCOLOR:%s\n", chat->dtcolor);
  for(i = 0; i < MAX_LINKS; i++)
  {
    if(*chat->links[i].url != '\0')
      fprintf(fp, "LINK:%s,%s,%s\n", chat->links[i].url,
        chat->links[i].label_ja, chat->links[i].label_en);
  }
  fclose(fp);
}
void import_ini(chat)
CHAT *chat;
{
  FILE *fp;
  char buf[1024];
  int i, n = 0;
  fp = fopen(EXPORT_INI, "r");
  if(fp == NULL) return;
  while(fgets(buf, sizeof(buf), fp) != NULL)
  {
    for(i=strlen(buf)-1;i>=0;i--)
      if(buf[i] >= '\0' && buf[i] <= ' ') buf[i] = '\0';
      else break;
    if(strncmp(buf, "TITLE:", 6) == 0)
      strncpy(chat->title, &buf[6], TITLE_LEN);
    else if(strncmp(buf, "TITLEIMAGE:", 11) == 0)
      strncpy(chat->titleimage, &buf[11], ADDR_LEN);
    else if(strncmp(buf, "BGCOLOR:", 8) == 0)
      strncpy(chat->bgcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "BGIMAGE1:", 9) == 0)
      strncpy(chat->bgimage1, &buf[9], ADDR_LEN);
    else if(strncmp(buf, "BGIMAGE2:", 9) == 0)
      strncpy(chat->bgimage2, &buf[9], ADDR_LEN);
    else if(strncmp(buf, "BGIMAGE3:", 9) == 0)
      strncpy(chat->bgimage3, &buf[9], ADDR_LEN);
    else if(strncmp(buf, "TXCOLOR:", 8) == 0)
      strncpy(chat->txcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "LICOLOR:", 8) == 0)
      strncpy(chat->licolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "VLCOLOR:", 8) == 0)
      strncpy(chat->vlcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "ALCOLOR:", 8) == 0)
      strncpy(chat->alcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "DNCOLOR:", 8) == 0)
      strncpy(chat->dncolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "DTCOLOR:", 8) == 0)
      strncpy(chat->dtcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "LINK:", 5) == 0)
    {
      if(n < MAX_LINKS)
      {
        char *jp, *en;
        if((jp = (char *)strchr(buf, ',')) == NULL) continue;
        *jp = '\0';
        jp++;
        if((en = (char *)strchr(jp, ',')) != NULL)
        {
          *en = '\0';
          en++;
        }
        else
        {
          en = jp;
        }
        strncpy(chat->links[n].url, &buf[5], ADDR_LEN);
        strncpy(chat->links[n].label_ja, jp, LABEL_LEN);
        strncpy(chat->links[n].label_en, en, LABEL_LEN);
        n++;
      }
    }
  }
  fclose(fp);
}
#endif /* EXPORT_INI */

#ifdef EXPORT_CHAT
void tab2spc(str)
char *str;
{
  while(*str != '\0')
  {
    if(*str == '\t') *str = ' ';
    str++;
  }
}

void export_chat(chat, mode)
CHAT *chat;
int mode;
{
  FILE *fp;
  int i;
  fp = fopen(EXPORT_CHAT, "w");
  if(fp != NULL)
  {
    for(i = 0; i < MAX_LINES; i++)
    {
      TEXT *txline = (TEXT *)gettextline(chat, i);
      char namecolor[COLOR_LEN];
      char name[NAME_LEN];
      char textcolor[COLOR_LEN];
      char message[MSG_SIZE];
      char atime[11];
      strcpy(namecolor, txline->namecolor);
      strcpy(name, txline->name);
      strcpy(textcolor, txline->textcolor);
      strcpy(message, txline->message);
      sprintf(atime, "%d", (int)txline->atime);
      tab2spc(name);
      tab2spc(namecolor);
      tab2spc(textcolor);
      tab2spc(message);
      fprintf(fp, "%s\t%s\t%s\t%s\t%s\n", namecolor, name, textcolor, message, atime);
    }
    fclose(fp);
  }
}
void import_chat(chat)
CHAT *chat;
{
  FILE *fp;
  int i;
  fp = fopen(EXPORT_CHAT, "r");
  if(fp != NULL)
  {
    for(i = 0; i < MAX_LINES; i++)
    {
      char tmp[COLOR_LEN+NAME_LEN+COLOR_LEN+MSG_SIZE+20];
      char *namecolor;
      char *name;
      char *textcolor;
      char *message;
      char *atime;
      int j = 0;
      if(fgets(tmp, sizeof(tmp), fp) == NULL) break;
      namecolor = &tmp[j];
      for(; j < sizeof(tmp) && !((tmp[j] >= 0x0)&&(tmp[j] < 0x20)); j++);
      tmp[j++] = '\0';
      name = &tmp[j];
      for(; j < sizeof(tmp) && !((tmp[j] >= 0x0)&&(tmp[j] < 0x20)); j++);
      tmp[j++] = '\0';
      textcolor = &tmp[j];
      for(; j < sizeof(tmp) && !((tmp[j] >= 0x0)&&(tmp[j] < 0x20)); j++);
      tmp[j++] = '\0';
      message = &tmp[j];
      for(; j < sizeof(tmp) && !((tmp[j] >= 0x0)&&(tmp[j] < 0x20)); j++);
      tmp[j++] = '\0';
      atime = &tmp[j];
      for(; j < sizeof(tmp) && !((tmp[j] >= 0x0)&&(tmp[j] < 0x20)); j++);
      tmp[j++] = '\0';
      
      strcpy(chat->text[MAX_LINES - 1 - i].namecolor, namecolor);
      strcpy(chat->text[MAX_LINES - 1 - i].name, name);
      strcpy(chat->text[MAX_LINES - 1 - i].textcolor, textcolor);
      strcpy(chat->text[MAX_LINES - 1 - i].message, message);
      chat->text[MAX_LINES - 1 - i].atime = atoi(atime);
    }
    chat->text_line = 0;
    fclose(fp);
  }
}
#endif /* EXPORT_CHAT */


int specialcmd(chat, message)
CHAT *chat;
char *message;
{
  int i, j;
  char arg1[1024], arg2[1024], arg3[1024];
  int argc, f = 0;

  if(*message == '#')
  {
    f = 1;
    message++;
  }
  if(*message != '<' && *message != '/' && *message != '\\') return(0);

  if(*message == '\\')
  {
    strcpy(arg1, "\\");
    argc = 1;
  }
  else
    argc = get_tagparam(message, 1024, arg1, arg2, arg3);

  if(*message == '/') itag = 1;
  scstrlwr(arg1);

  if(strcmp(arg1, "who") == 0)
  {
    userproperty = 1;
    return(1);
  }
  else if(strcmp(arg1, "cancel") == 0)
  {
    int n = 1;
    if(argc == 1)
    {
      strncpy(arg2, name, sizeof(arg2));
      if(userid > 0 && chat->user[userid].count > 0)
        chat->user[userid].count --;
    }
    else if(permission < 2) goto cmd_error;

    if(argc == 3)
      n = atoi(arg3);

    for(i = 1; i < MAX_LINES; i++)
    {
      j = (MAX_LINES+chat->text_line-i)%MAX_LINES;
      if((strcmp(chat->text[j].name, arg2) == 0) &&
         (*chat->text[j].message != '\0') )
      {
        *chat->text[j].message = '\0';
        if(--n > 0) continue;
        break;
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "delete") == 0)
  {
    if(permission < 2 || argc > 2) goto cmd_error;
    if(argc == 2)
    {
      int i;
      i = atoi(arg2);
      *chat->text[i].message = '\0';
      chat->text[i].atime = 0;
    }
    else
    {
      deletebtn = 1;
    }
    return(1);
  }
  else if(strcmp(arg1, "log") == 0)
  {
    int log;
    if(permission < 1 || argc < 2) goto cmd_error;
    log = atoi(arg2);
    if(log < 1 || log > MAX_LINES) goto cmd_error;
    if(isIMODE(ostype))
    {
      ilog = log;
      idate = 1;
    }
    else
      chat->user[userid].maxlines = log;
    maxlines = log;
    return(1);
  }
  else if(strcmp(arg1, "slog") == 0)
  {
    int log;
    if(permission < 1 || argc < 2) goto cmd_error;
    log = atoi(arg2);
    if(log < 1 || log > MAX_LINES) goto cmd_error;
    ilog = log;
    if(isIMODE(ostype))
      maxlines = log;
    return(1);
  }
  else if(strcmp(arg1, "color") == 0)
  {
    if(argc < 2)
    {
      *chat->user[userid].namecolor = '\0';
      *chat->user[userid].textcolor = '\0';
    }
    else
    {
      if(strcmp(arg2, "-") != 0)
      {
        if((arg2[0] == '=') && (argc == 2))
        {
          int i;
          for(i = 0; i < MAX_USERS; i++)
          {
            if(strcmp(chat->user[i].name, &arg2[1]) == 0)
            {
               strncpy(chat->user[userid].namecolor, chat->user[i].namecolor, COLOR_LEN);
               strncpy(chat->user[userid].textcolor, chat->user[i].textcolor, COLOR_LEN);
               break;
            }
          }
          if(i >= MAX_USERS) goto cmd_error;
        }
        else
        {
          if(isvalidcolor(arg2, def_bgcolor) != 0) goto cmd_error;
          strncpy(chat->user[userid].namecolor, arg2, COLOR_LEN);
        }
      }
    }
    if(argc == 3)
    {
      if(isvalidcolor(arg3, def_bgcolor) != 0) goto cmd_error;
      strncpy(chat->user[userid].textcolor, arg3, COLOR_LEN);
    }
    return(1);
  }
  else if(strcmp(arg1, "lock") == 0)
  {
    int level;
    if(permission < 2 || argc < 2) goto cmd_error;
    level = atoi(arg2);
    switch(level)
    {
    case 0:
      chat->url[0] = CHAT_LOCK_PUBLIC;
      break;
    case 1:
      chat->url[0] = CHAT_LOCK_REGISTED;
      break;
    case 2:
      chat->url[0] = CHAT_LOCK_PERMIT;
      break;
    case 3:
      chat->url[0] = CHAT_LOCK_CURRENT;
      break;
    default:
      goto cmd_error;
    }
    return(1);
  }
  else if(strcmp(arg1, "deny") == 0 || strcmp(arg1, "undeny") == 0)
  {
    IP dh;
    int tim = REFUSETIME;
    int ds = -1;
    int dm = 1;
    if(arg1[0] == 'u') dm = 0;
    if(permission < 2 || argc < 2) goto cmd_error;
    dh = IPstr2IP(arg2);
    if(argc == 3)
      tim = atoi(arg3);
    if(dh != -1)
    {
      /* (un)deny by IP */
      ds = denyhost(chat, dh, tim, "", dm);
    }
    else
    {
      /* (un)deny by name */
      for(i=1;i<MAX_USERS;i++)
      {
        if(strcmp(chat->user[i].name, arg2) == 0)
        {
          ds = denyhost(chat, chat->user[i].haddressip, tim, arg2, dm);
          if(ds == 0 && dm == 1)
            chat->user[i].login = 2;
        }
      }
    }
    if(ds == -1)
    {
      sendmessage(alertmsg, sizeof(alertmsg), MSG(M_COMMANDFAIL), MT_ERROR);
      return(1);
    }
    return(1);
  }
  else if(strcmp(arg1, "reset") == 0)
  {
    if(argc == 1) /* clear password */
    {
        *chat->user[userid].password = '\0';
        chat->user[userid].mid = userid;
        if(chat->user[userid].permit >= 2)
          chat->user[userid].permit = 1;
        export_user(EXPORT_USER, &chat->user[userid], 0);
        userreseted = 1;
        return(1);
    }
    if(permission < 2 || argc < 2) goto cmd_error;
    for(i = 1; i < MAX_USERS; i++)
    {
      if(strcmp(chat->user[i].name, arg2) == 0 &&
         chat->user[i].permit >= 0)
      {
        if(chat->user[i].permit > 1)
        {
          chat->user[i].permit = 1;
        }
        else
        {
          chat->user[i].permit = 0;
          if(*chat->user[i].password == '\0')
          {
            *chat->user[i].name = '\0';
            chat->user[i].mid = userid;
            return(1);
          }
          *chat->user[i].password = '\0';
        }
        chat->user[i].mid = userid;
        export_user(EXPORT_USER, &chat->user[i], 0);
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "permit") == 0)
  {
    if(permission < 1 || argc < 2) goto cmd_error;
    for(i = 1; i < MAX_USERS; i++)
    {
      if(strcmp(chat->user[i].name, arg2) == 0)
      {
        if(*chat->user[i].password != '\0')
        {
          if(chat->user[i].permit < permission)
          {
            chat->user[i].permit ++;
            chat->user[i].mid = userid;
            export_user(EXPORT_USER, &chat->user[i], 0);
          }
          break;
        }
        else /* non passworded user */
        {
          if((chat->user[i].login == 1) &&
             (get_atime(chat, i) > now - AUTOLOGOUT))
          {
            chat->user[i].permit = 1;
            chat->user[i].mid = userid;
          }
        }
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "crash") == 0)
  {
    if(permission < 1 || argc < 2) goto cmd_error;
    for(i = 1; i < MAX_USERS; i++)
    {
      if(strcmp(chat->user[i].name, arg2) == 0)
      {
        if((permission >= 2) || (chat->user[i].permit < permission))
          SET_FLAG(chat->user[i].flags, USER_CRASH);
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "uncrash") == 0)
  {
    if(permission < 1 || argc < 2) goto cmd_error;
    for(i = 1; i < MAX_USERS; i++)
    {
      if(strcmp(chat->user[i].name, arg2) == 0)
      {
        if((permission >= 2) || (chat->user[i].permit < permission))
          RESET_FLAG(chat->user[i].flags, USER_CRASH);
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "topic") == 0)
  {
    int i;
    char *p = &message[6];
    if(permission < 1) goto cmd_error;
    while(*p == ' ') p++;
#ifndef NOROMAKANA
    if(*p == '\\')
      roma2kana(arg2, ++p, sizeof(arg2));
    else
#endif /* NOROMAKANA */
      strncpy(arg2, p, sizeof(arg2));
    if((i = strlen(arg2)) > 0)
      if((message[0] == '<') && (arg2[i - 1] == '>')) arg2[i - 1] = '\0';
    strncpy(chat->topic.name, name, NAME_LEN);
    strncpy(chat->topic.message, arg2, MSG_SIZE);
    chat->topic.atime = now;
    return(1);
  }
  else if(strcmp(arg1,"alert") == 0)
  {
    char msg[MEMO_SIZE];
    int i, j=0;
    if(argc < 3) goto cmd_error;
    if(strcmp(arg2, managerid) == 0)
    {
      sendmessage(alertmsg, sizeof(alertmsg), MSG(M_CANNOTSEND), MT_ERROR);
      return(1);
    }
    /* for(i=0;arg3[i] != '\0';i++)
      if(arg3[i] == '"') arg3[i] = ' ';
    */
    printdate(msg, now);
    strcat(msg, "\\n");
    strcat(msg, name);
    strcat(msg, "> ");
    i = strlen(msg);
    strncpy(&msg[i], arg3, sizeof(msg) - i);
    for(i = 1; i < MAX_USERS; i++)
    {
      if(strcmp(chat->user[i].name, arg2) == 0 ||
         (ismanager() && (*arg2 == '*')))
      {
        if(!IS_FLAG(chat->user[i].flags, USER_REFUSE))
        {
          if(!sendmessage(chat->user[i].memo, MEMO_SIZE, msg, MT_ALERT))
          {
            sendmessage(alertmsg, sizeof(alertmsg),MSG(M_CANNOTSEND),MT_ERROR);
            break;
          }
          j ++;
        }
      }
    }
    if(j == 0)
    {
      sendmessage(alertmsg, sizeof(alertmsg), MSG(M_CANNOTSEND), MT_ERROR);
    }
    chat->user[userid].altime = now;
    return(1);
  }
  else if(strcmp(arg1, "refresh") == 0)
  {
    int ref;
    ref = atoi(arg2);
    if(ref < 0 || ref > 120) goto cmd_error;
    chat->user[userid].refresh = ref;
    return(1);
  }
  else if(strcmp(arg1, "skiptitle") == 0)
  {
    if(argc == 2 && strcmp(arg2, "off") == 0)
    {
      RESET_FLAG(chat->user[userid].flags, USER_SKIPTITLE);
      skiptitle = 0;
    }
    else
    {
      SET_FLAG(chat->user[userid].flags, USER_SKIPTITLE);
      skiptitle = 1;
    }
    return(1);
  }
  else if(strcmp(arg1, "refuse") == 0)
  {
    if(argc == 2 && strcmp(arg2, "off") == 0)
    {
      RESET_FLAG(chat->user[userid].flags, USER_REFUSE);
      refuse = 0;
    }
    else
    {
      SET_FLAG(chat->user[userid].flags, USER_REFUSE);
      refuse = 1;
    }
    return(1);
  }
  else if(strcmp(arg1, "hr") == 0)
  {
    if(argc == 2 && strcmp(arg2, "off") == 0)
    {
      RESET_FLAG(chat->user[userid].flags, USER_TEXTHR);
      texthr = 0;
    }
    else
    {
      SET_FLAG(chat->user[userid].flags, USER_TEXTHR);
      texthr = 1;
    }
    return(1);
  }
  else if(strcmp(arg1, "su") == 0)
  {
    char tmp[256];
    sprintf(tmp, "%s/%s", managerid, managerpass);
    if(*password != '\0' &&
       strcmp(arg2, tmp) == 0 && userid >= 0)
    {
      chat->user[userid].permit = 3;
      chat->user[userid].mid = userid;
      export_user(EXPORT_USER, &chat->user[userid], 0);
      return(1);
    }
    goto cmd_error;
  }
  else if(strcmp(arg1, "set") == 0)
  {
    if(permission < 3) goto cmd_error;
    if(argc < 2) goto cmd_error;
    if(argc == 2)
    {
        if(strcmp(arg2, "title") == 0)
            strncpy(chat->title, MSG(M_TITLE), TITLE_LEN);
        else if(strcmp(arg2, "titleimage") == 0)
            strcpy(chat->titleimage, "");
        else if(strcmp(arg2, "bgcolor") == 0)
            strcpy(chat->bgcolor, "");
        else if(strcmp(arg2, "bgimage") == 0)
            strcpy(chat->bgimage1, "");
        else if(strcmp(arg2, "txcolor") == 0)
            strcpy(chat->txcolor, "");
        else goto cmd_error;
    }
    else if(argc == 3)
    {
        if(strcmp(arg2, "title") == 0)
            strncpy(chat->title, arg3, TITLE_LEN);
        else if(strcmp(arg2, "titleimage") == 0)
            strncpy(chat->titleimage, arg3, ADDR_LEN);
        else if(strcmp(arg2, "bgcolor") == 0)
            strncpy(chat->bgcolor, arg3, COLOR_LEN);
        else if(strcmp(arg2, "bgimage") == 0)
            strncpy(chat->bgimage1, arg3, ADDR_LEN);
        else if(strcmp(arg2, "txcolor") == 0)
            strncpy(chat->txcolor, arg3, COLOR_LEN);
        else goto cmd_error;
    }
#ifdef EXPORT_INI
    export_ini(chat);
#endif
    return(1);
  }
  else if(strcmp(arg1, "export") == 0)
  {
    if(permission < 3) goto cmd_error;
#ifdef EXPORT_USER
    export_users(chat);
#endif
#ifdef EXPORT_BBS
    export_bbss(chat);
#endif
#ifdef EXPORT_CHAT
    export_chat(chat, 0);
#endif
    return(1);
  }
  else if(strcmp(arg1, "import") == 0)
  {
    if(permission < 3) goto cmd_error;
#ifdef EXPORT_USER
    import_users(chat);
#endif
#ifdef EXPORT_BBS
    import_bbss(chat);
#endif
#ifdef EXPORT_CHAT
    import_chat(chat);
#endif
    return(1);
  }
#ifndef NOROMAKANA
  else if(strcmp(arg1, "r") == 0 || strcmp(arg1, "\\") == 0)
  {
    char *p, kana[MSG_SIZE] = "";
    if(*arg1 == '\\')
      roma2kana(kana, &message[1], sizeof(kana));
    else if((p = (char *)strchr(message, '>')) != NULL)
      roma2kana(kana, ++p, sizeof(kana));
    strncpy(message, kana, MSG_SIZE);
  }
#endif /* NOROMAKANA */
  if(f == 0)
    return(0);
  
cmd_error:
  sendmessage(alertmsg, sizeof(alertmsg), MSG(M_COMMANDFAIL), MT_ERROR);
  return(1);
}

int specialbbscmd(chat, message)
CHAT *chat;
char *message;
{
  char arg1[1024], arg2[1024], arg3[1024];
  int argc;

  if(*message != '<' && *message != '/') return(0);

  argc = get_tagparam(message, 1024, arg1, arg2, arg3);
  if(strcmp(arg1, "cancel") == 0)
  {
    int i, j;
    for(i=1;i<=MAX_BBS;i++)
    {
      j = (MAX_BBS+chat->bbs_line-i)%MAX_BBS;
      if((strcmp(chat->bbs[j].name, name) == 0) &&
         (*chat->bbs[j].message != '\0') )
      {
        strncpy(message, chat->bbs[j].message, MSG_SIZE);
        *chat->bbs[j].message = '\0';
        chat->bbs[j].atime = 0;
        break;
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "delete") == 0)
  {
    int i;
    i = atoi(arg2);
    strncpy(message, chat->bbs[i].message, MSG_SIZE);
    message[strlen(chat->bbs[i].message)] = '\0';
    *chat->bbs[i].message = '\0';
    chat->bbs[i].atime = 0;
    return(1);
  }
#ifndef NOROMAKANA
  else if(strcmp(arg1, "r") == 0 || strcmp(arg1, "\\") == 0 )
  {
    char *p, kana[MSG_SIZE] = "";
    if(*arg1 == '\\')
      roma2kana(kana, &message[1], sizeof(kana));
    if((p = (char *)strchr(message, '>')) != NULL)
      roma2kana(kana, ++p, sizeof(kana));
    strncpy(message, kana, MSG_SIZE);
  }
#endif /* NOROMAKANA */
  return(0);
}

int shmid;
#define SHM_MODE 0666

#ifdef USE_SEM
int semid;
#define SEM_MODE SHM_MODE
static struct sembuf semops[1];
#ifndef SEM_KEY
#define SEM_KEY SHM_KEY
#endif

#ifdef DEF_SEMUN
union semun {
  int val;
  struct semid_ds *buf;
  ushort *array;
} semv;
#else /* !DEF_SEMUN */
union semun semv;
#endif /* DEF_SEMUN */

#endif /* USE_SEM */

CHAT *get_chat()
{
  CHAT *chat;
#ifdef USE_SEM
  if((semid = semget(SEM_KEY, 1, SEM_MODE)) == EOF)
  {
    semv.val = 1;
    if((semid = semget(SEM_KEY, 1, SEM_MODE|IPC_CREAT)) == EOF)
    {
      scerr(MSG(M_CANNOTUSE));
    }
    semctl(semid, 0, SETVAL, semv); /* unlock */
  }
#endif /* USE_SEM */
  if((shmid=shmget(SHM_KEY, sizeof(CHAT), SHM_MODE)) == EOF)
  {
    int retry = 0;
retry_get:
    if((shmid=shmget(SHM_KEY, sizeof(CHAT), SHM_MODE|IPC_CREAT)) == EOF)
    {
      if((shmid=shmget(SHM_KEY, 0, SHM_MODE)) == EOF)
      {
        scerr(MSG(M_CANNOTUSE));
      }
      if(shmctl(shmid, IPC_RMID, 0) == EOF)
      {
        scerr(MSG(M_CANNOTUSE));
      }
      if(retry == 0)
      {
        retry++;
        goto retry_get;
      }
      scerr(MSG(M_CANNOTUSE));
    }
    if((int)(chat = (CHAT *)shmat(shmid, 0, 0)) == EOF)
    {
      scerr(MSG(M_CANNOTUSE));
    }
    init_memory(chat); /* initialize work area */
    shmdt(chat);
    chat = NULL;
  }
  if((int)(chat = (CHAT *)shmat(shmid, 0, 0)) == EOF)
  {
    scerr(MSG(M_CANNOTUSE));
  }
  if(chat->version != memversion)
    reset_chat(chat);
  return(chat);
}

void release_chat(chat)
CHAT *chat;
{
  shmdt(chat);
  chat = NULL;
}

#ifdef WIN32
void lock_chat(chat)
CHAT *chat;
{ 
}
void unlock_chat(chat)
CHAT *chat;
{
}
#else /* WIN32 */
void lock_chat(chat)
CHAT *chat;
{ 
  signal(SIGALRM, sigalm);
  alarm(TIMEOUT);
#ifdef USE_SEM
  semops[0].sem_num = 0;
  semops[0].sem_op = -1; /* Lock */
  semops[0].sem_flg = 0; /* 0: wait  IPC_NOWAIT: no wait */
  semop(semid, semops, 1);
#else  /* !USE_SEM */
  while(chat->memlock);
  chat->memlock = 1;
#endif /* USE_SEM */
}

void unlock_chat(chat)
CHAT *chat;
{
  signal(SIGALRM, SIG_IGN);
#ifdef USE_SEM
  semops[0].sem_num = 0;
  semops[0].sem_op = 1; /* Unlock */
  semops[0].sem_flg = 0;
  semop(semid, semops, 1);
#else  /* !USE_SEM */
  chat->memlock = 0;
#endif /* USE_SEM */
}
#endif /* WIN32 */

void reset_chat(chat)
CHAT *chat;
{
  if((int)shmctl(shmid, IPC_RMID, 0) == EOF)
  {
    scerr(MSG(M_CANNOTUSE));
  }
#ifdef USE_SEM
  semv.val = 0;
  if((int)semctl(semid, 0, IPC_RMID, semv) == EOF)
  {
    scerr(MSG(M_CANNOTUSE));
  }
#endif /* USE_SEM */
#ifdef EXPORT_CHAT
/*
  {
    FILE *fp;
    if((fp = fopen(EXPORT_CHAT, "w")) != NULL) fclose(fp);
  }
*/
#endif /* EXPORT_CHAT */
  
  scerr(MSG(M_INITIALIZED));
}

int get_lockmode(chat)
CHAT *chat;
{
  if(ismanager()) return 0;
  return(chat->url[0]);
}

CHATMODE get_mode(mode)
char *mode;
{
  if(strcmp(mode, "entry") == 0)
    return(m_entry);
  else if(strcmp(mode, "input") == 0)
    return(m_input);
  else if(strcmp(mode, "chattext") == 0)
    return(m_chattext);
  else if(strcmp(mode, "bbs") == 0)
    return(m_bbs);
  else if(strcmp(mode, "logout") == 0)
    return(m_logout);
  else if(strcmp(mode, "customize") == 0)
    return(m_customize);
  else if(strcmp(mode, "reset") == 0)
    return(m_reset);
  else if(strcmp(mode, "debug") == 0)
    return(m_debug);
  else if(strcmp(mode, "doc") == 0)
    return(m_doc);
  else if(strcmp(mode, "event") == 0)
    return(m_event);
  return(m_init);
}

#define IFSETSPARAM(x,y) \
  if(strncmp(text,y "=",sizeof(y))==0 && \
     strlen(text)>sizeof(y)) \
      strncpy((x), &text[sizeof(y)], sizeof(x))

#define IFCATSPARAM(x,y) \
  if(strncmp(text,y "=",sizeof(y))==0 && \
     strlen(text)>sizeof(y)) \
      strncat((x), &text[sizeof(y)], sizeof(x) - strlen(x))

#define IFSETIPARAM(x,y) \
  if(strncmp(text,y "=",sizeof(y))==0 && \
     strlen(text)>sizeof(y)) \
      x = atoi(&text[sizeof(y)])

#define IFSETBPARAM(x,y) \
  if(strncmp(text,y,sizeof(y)-1)==0 && \
     strlen(text)>=sizeof(y)-1) \
      x = 1

int main(argc, argv)
int argc;
char **argv;
{
  char *param;
  char text[MSG_SIZE];
  char *src, *proxy;
  int i;
  int chatlines = 0;
  time_t logintime = 0;
  IP httpfor = 0;
  char alang[256];

  if((argc > 1) && (strcmp(argv[1], "-export") == 0))
  {
    chat = get_chat();
#ifdef EXPORT_USER
    export_users(chat);
    printf("export users: %s\n", EXPORT_USER);
#endif
#ifdef EXPORT_BBS
    export_bbss(chat);
    printf("export bbs: %s\n", EXPORT_BBS);
#endif
#ifdef EXPORT_CHAT
    export_chat(chat, 0);
    printf("export chat: %s\n", EXPORT_CHAT);
#endif
    chat_exit(0);
  }

  signal(SIGTERM, sigterm);
  signal(SIGINT, sigterm);
  signal(SIGSEGV, sigterm);

#ifdef COOKIE
  strncpy(cookie, get_env("HTTP_COOKIE"), sizeof(cookie));
  if(*cookie != '\0')
  {
    char *cp;
    cp = strstr(cookie, "chatname=");
    if(cp != 0)
    {
      cp += 9;
      for(i = 0; (i < sizeof(c_chatname)) && (*cp != '\0') && (*cp != ' ') && (*cp != ';'); i++, cp++)
      {
        c_chatname[i] = *cp;
      }
      c_chatname[i] = '\0';
    }
    cp = strstr(cookie, "refresh=");
    if(cp != 0)
    {
      cp += 8;
      for(i = 0; (i < sizeof(c_refresh)) && (*cp != '\0') && (*cp != ' ') && (*cp != ';'); i++, cp++)
      {
        c_refresh[i] = *cp;
      }
      c_refresh[i] = '\0';
    }
    cp = strstr(cookie, "autoclear=");
    if(cp != 0)
    {
      cp += 10;
      for(i = 0; (i < sizeof(c_autoclear)) && (*cp != '\0') && (*cp != ' ') && (*cp != ';'); i++, cp++)
      {
        c_autoclear[i] = *cp;
      }
      c_autoclear[i] = '\0';
    }
  }
#endif /* COOKIE */
  strncpy(agent, get_env("HTTP_USER_AGENT"), sizeof(agent));
  strncpy(alang, get_env("HTTP_ACCEPT_LANGUAGE"), sizeof(alang));
#ifdef HTTPD_HOSTNAMELOOKUP
  strncpy(haddress, get_env("REMOTE_HOST"), sizeof(haddress));
#endif /* HTTPD_HOSTNAMELOOKUP */
  if(*(proxy = get_env("HTTP_X_FORWARDED_FOR")) != '\0') /* via Squid */
  {
    if(isIP(proxy)) httpfor = IPstr2IP(proxy);
  }
  else if(*(proxy = get_env("HTTP_FROM")) != '\0') /* via DeleGate4.xx */
  {
    proxy += 2;
    if(isIP(proxy)) httpfor = IPstr2IP(proxy);
  }
  if(httpfor != 0)
  {
    int len = strlen(haddress), maxlen = sizeof(haddress);
    /* private nets (RFC 1918) */
    if(!isprivateIP(httpfor))
    {
      if(len < maxlen - 3)
      {
        strcpy(&haddress[len], "->");
        strncpy(&haddress[len+2], proxy, maxlen - len - 2);
      }
    }
  }
    
  haddressip = IPstr2IP(get_env("REMOTE_ADDR"));
  /* Agent Type */
  if(strstr(agent, "Mozilla") != NULL)
  {
    OUTPUT_JCODE = OC_SJIS;
    if(strstr(agent, "/2.02") != NULL)
      oldmozilla = 1;
    else if(strstr(agent, "Gecko/") != NULL)
      mozilla6 = 1;
    if(strstr(agent, "; U)") != NULL)
    {
      strcpy(lang, "en");
    }
    if(alang[0] != '\0') {
      if(strstr(alang, "ja") != NULL)
      {
        strcpy(lang, "ja");
      }
      else if(strstr(alang, "en") != NULL)
      {
        strcpy(lang, "en");
      }
      else if(strstr(alang, "it") != NULL)
      {
        strcpy(lang, "it");
      }
      else if(strstr(alang, "zh") != NULL)
      {
        strcpy(lang, "zh");
      }
      else if(strstr(alang, "ko") != NULL)
      {
        strcpy(lang, "kr");
      }
      else
      {
        strcpy(lang, "en");
      }
    }
  }
  else if(strstr(agent, "DoCoMo/") != NULL)
  {
    OUTPUT_JCODE = OC_SJIS;
    ostype = OS_IMODE;
  }
  else if(strstr(agent, "J-PHONE/") != NULL)
  {
    strcpy(cgimtd, "GET");
    OUTPUT_JCODE = OC_SJIS;
    ostype = OS_JPHONE;
  }
  else if(strstr(agent, "UP.Browser/") != NULL)
  /*if(strstr(agent, "D502i") != NULL)*/
  {
    strcpy(cgimtd, "GET");
    OUTPUT_JCODE = OC_SJIS;
    ostype = OS_AU;
  }
  else if(strstr(agent, "sharp pda browser") != NULL
    || strstr(agent, "sharp wd browser") != NULL
    || strstr(agent, "Plala") != NULL
    || strstr(agent, "JustView") != NULL
    || strstr(agent, "Aplix_SANYO_browser") != NULL
    || strstr(agent, "Foliage iBrowser") != NULL
    )
  {
    OUTPUT_JCODE = OC_SJIS;
  }

  if(strstr(agent, "Macintosh") != NULL)
  {
    ostype = OS_MAC;
  }

  if(*agent == '\0')
  { /* J-SkyWeb? */
    strncpy(agent, get_env("HTTP_X_JPHONE_MSNAME"), sizeof(agent));
    strcpy(cgimtd, "GET");
    OUTPUT_JCODE = OC_SJIS;
    ostype = OS_JPHONE;
  }

  now = time(NULL);
  nowtm = localtime(&now);
  chat_mode = m_init;

  chat = get_chat();
  if(chat == NULL)
  {
    scerr(MSG(M_CANNOTUSE));
  }
  strncpy(def_title, chat->title, sizeof(def_title));
  strncpy(def_bgcolor, chat->bgcolor, sizeof(def_bgcolor));
  strncpy(def_bgimage, chat->bgimage1, sizeof(def_bgimage));
  strncpy(def_txcolor, chat->txcolor, sizeof(def_txcolor));
  strncpy(def_licolor, chat->licolor, sizeof(def_licolor));
  strncpy(def_vlcolor, chat->vlcolor, sizeof(def_vlcolor));
  strncpy(def_alcolor, chat->alcolor, sizeof(def_alcolor));
  strncpy(def_dncolor, chat->dncolor, sizeof(def_dncolor));
  strncpy(def_dtcolor, chat->dtcolor, sizeof(def_dtcolor));
  SETBODY(body, def_bgcolor, def_bgimage, def_txcolor, def_licolor, def_vlcolor, def_alcolor);
  if(*def_title == '\0') title = MSG(M_TITLE);
  else                   title = def_title;

  if((param = getcgiarg()) == NULL)
    exit(1);

  src = param;
  while(*src != '\0')
  {
    src = getparam(text, sizeof(text), src);
    if(strncmp(text,"URL=",4) == 0)
    {
      print_forward(&text[4]);
      chat_exit(0);
    }
    IFSETSPARAM(message, "chat");
    else IFSETIPARAM(refresh, "refresh");
    else IFSETIPARAM(maxlines, "maxlines");
    else IFSETIPARAM(bbsline, "bbsline");
    else IFSETSPARAM(mytitle, "mytitle");
    else IFSETSPARAM(bgcolor, "bgcolor");
    else IFSETSPARAM(txcolor, "txcolor");
    else IFSETSPARAM(rawpassword, "pass");
    else IFSETSPARAM(cpassword, "password");
    else IFSETSPARAM(address, "address");
    else IFSETSPARAM(referer, "referer");
    else IFSETSPARAM(def_title, "def_title");
    else IFSETSPARAM(def_bgcolor, "def_bgcolor");
    else IFSETSPARAM(def_txcolor, "def_txcolor");
    else IFSETSPARAM(def_bgimage,  "def_bgimage");
    else IFSETSPARAM(def_licolor, "def_licolor");
    else IFSETSPARAM(def_vlcolor, "def_vlcolor");
    else IFSETSPARAM(def_alcolor, "def_alcolor");
    else IFSETSPARAM(def_dncolor, "def_dncolor");
    else IFSETSPARAM(def_dtcolor, "def_dtcolor");
    else IFSETSPARAM(mode, "mode");
    else IFSETSPARAM(lang, "mlang");
    else IFSETSPARAM(jcode, "jcode");
    else IFSETSPARAM(secure, "secure");
    else IFSETIPARAM(room, "room");
    else if(strncmp(text, "link_url=", 9) == 0)
    {
      if(link_items < MAX_LINKS)
        strncpy(link_url[link_items], &text[9], ADDR_LEN);
    }
    else if(strncmp(text, "link_label=", 11) == 0)
    {
      if(link_items < MAX_LINKS)
      {
        strncpy(link_label[link_items], &text[11], LABEL_LEN);
        link_items++;
      }
    }
    else if(strncmp(text, "name=", 5) == 0 && strlen(text) > 5)
    {
      char inputname[NAME_LEN+PASS_LEN+2];
      char *n = &text[5], *p = NULL;
      while((*n == ' ') && (*n != '\0')) n++;
      strncpy(inputname, n, sizeof(inputname));
      for(i=strlen(inputname)-1;i>0;i--)
        if(inputname[i]>=0&&inputname[i]<=0x20) inputname[i] = '\0';
      if((p = (char *)strchr(inputname, '/')) != NULL) /* with a password */
      {
        strncpy(password, p+1, sizeof(password));
        encode(cpassword, password);
        *p = '\0';
        strncpy(name, inputname, sizeof(name));
        if(*name != '\0') *address = '\0';
      }
      else
      {
        strncpy(name, inputname, sizeof(name));
      }
      killtag(name);
      if(*name == '\0' && *password != '\0') /* password only */
      {
        if(get_userbypassword(name, address, password) == -1 ||
           strcmp(password, PASSWORD) == 0)
        {
          password[0] = '\0';
          address[0] = '\0';
          name[0] = '\0';
          sendmessage(alertmsg, sizeof(alertmsg), MSG(M_INVALIDPASS), MT_ERROR);
          back = 1;
        }
      }
    }
    else IFSETBPARAM(autoclear, "autoclear");
    else IFSETIPARAM(autoreload, "auto");
    else IFSETIPARAM(slog, "slog");
    else IFSETBPARAM(newtext, "newtext");
    else IFSETBPARAM(noframe, "noframe");
    else IFSETBPARAM(imode, "imode");
  }
  if(*rawpassword != '\0') encode(cpassword, rawpassword);
  decode(password, cpassword);
  killtag(address);
  chat_mode = get_mode(mode);
  if(*bgcolor != '\0')
    body.bgcolor = bgcolor;
  if(*txcolor != '\0')
    body.txcolor = txcolor;
  if(*lang != '\0')
  {
    int lg;
    lg = get_LANG(lang);
    if(lg != LANG) CONV = 1;
    LANG = lg;
  }

  if(imode)
  {
    ostype = OS_VIMODE;
    noframe = 1;
  }
  if(isIMODE(ostype))
  {
    newtext = 1;
  }
  if(*jcode != '\0')
  {
    if(strcmp(jcode, "jis") == 0)
      OUTPUT_JCODE = OC_JIS;
    else if(strcmp(jcode, "sjis") == 0)
      OUTPUT_JCODE = OC_SJIS;
    else if(strcmp(jcode, "euc") == 0)
      OUTPUT_JCODE = OC_EUC;
    else if(strcmp(jcode, "xbm") == 0)
      OUTPUT_JCODE = OC_XBM;
  }

  /* check deny hosts */
  if(*name != '\0' && !ismanager() && isdenyhost(chat, haddressip))
  {
    lock_chat(chat);
    userid = setuserinfo(chat);
    if(userid >= 0)
    {
      if(chat->user[userid].permit < 2)
        chat->user[userid].login = 2;
    }
    unlock_chat(chat);
    if(chat->user[userid].permit < 2)
      scerr(MSG(M_CANNOTUSE));
  }

  if(*name == '\0') /* entry form */
  {
    chat_mode = m_init;
  }

  switch(chat_mode)
  {
  case m_entry:
    if(get_lockmode(chat) == CHAT_LOCK_CURRENT)
      scerr(MSG(M_CANNOTUSE));
    lock_chat(chat);
    userid = setuserinfo(chat);
    unlock_chat(chat);
    if(!(noframe || isIMODE(ostype)))
    {
      release_chat(chat);
      enter_base();
    }
    chat_mode = m_chattext;
    break;
  case m_input:
    input_form(chat);
    break;
  case m_event:
    event_form(chat);
    break;
  case m_bbs:
  case m_logout:
    if(noframe || isIMODE(ostype))
    {
      lock_chat(chat);
      userid = setuserinfo(chat);
      unlock_chat(chat);
    }
    break;
  case m_reset:
    if(!ismanager())
      scerr(MSG(M_CANNOTUSE));
    reset_chat(chat);
    break;
  case m_debug:
    if(!ismanager())
      scerr(MSG(M_CANNOTUSE));
    printdebug(chat);
    chat_exit(0);
    break;
  case m_customize:
    if(ismanager())
    {
      lock_chat(chat);
      if(link_items == 0)
      {
        strncpy(chat->title, def_title, sizeof(def_title));
        strncpy(chat->bgcolor, def_bgcolor, sizeof(def_bgcolor));
        strncpy(chat->txcolor, def_txcolor, sizeof(def_txcolor));
        strncpy(chat->licolor, def_licolor, sizeof(def_licolor));
        strncpy(chat->vlcolor, def_vlcolor, sizeof(def_vlcolor));
        strncpy(chat->alcolor, def_alcolor, sizeof(def_alcolor));
        strncpy(chat->dncolor, def_dncolor, sizeof(def_dncolor));
        strncpy(chat->dtcolor, def_dtcolor, sizeof(def_dtcolor));
      }
      else
      {
        int i;
        for(i = 0; i < link_items; i++)
        {
          strncpy(chat->links[i].url, link_url[i], ADDR_LEN);
          strncpy(chat->links[i].label_ja, link_label[i], LABEL_LEN);
        }
      }
#ifdef EXPORT_INI
      export_ini(chat);
#endif
      unlock_chat(chat);
    }
    release_chat(chat);
    enter_base();
    break;
  case m_chattext:
    if(autoreload == 1 || cpassword[0] == '\0' || cpassword[1] == '[')
    {
      if(noframe || isIMODE(ostype))
      {
        lock_chat(chat);
        userid = setuserinfo(chat);
        unlock_chat(chat);
      }
      break;
    }
  case m_init:
  default:
    entry_form(chat);
    break;
  }

  local = (CHAT *)malloc(sizeof(CHAT));

  lock_chat(chat);

  userid = setuserinfo(chat);
  permission = chat->user[userid].permit;

  if(chat_mode == m_logout)
  {
    unlock_chat(chat);
    if(loginstat == 0)
      entry_form(chat);
    sleep(LOGOUTWAIT);
    lock_chat(chat);
    logintime = chat->user[userid].logintime;
  }

  if(chat_mode == m_bbs)
  {
    if(*message != '\0' && permission > 0)
    {
      if(specialbbscmd(chat, message) == 0)
      {
        int bbs_line = chat->bbs_line%MAX_BBS;
        *(chat->bbs[bbs_line].message) = '\0';
        chat->bbs[bbs_line].atime = now;
        strncpy(chat->bbs[bbs_line].name, name, NAME_LEN);
        strncpy(chat->bbs[bbs_line].namecolor, namecolor, COLOR_LEN);
        strncpy(chat->bbs[bbs_line].textcolor, textcolor, COLOR_LEN);
        strncpy(chat->bbs[bbs_line].message, message, MSG_SIZE);
        if(strcmp(secure, "on") == 0)
          SET_FLAG(chat->bbs[bbs_line].flags, SECURE_BBS);
        else
          RESET_FLAG(chat->bbs[bbs_line].flags, SECURE_BBS);
        chat->bbs_line++;
        *message = '\0';
      }
      export_bbss(chat);
    }
    lastbtime = chat->user[userid].btime;
    chat->user[userid].btime = now;
    if(*(chat->user[userid].password) != '\0')
      export_user(EXPORT_USER, &chat->user[userid], 0);
    memcpy(local, chat, sizeof(CHAT));

    unlock_chat(chat);
    release_chat(chat);

    printbbs(local, bbsline);
    chat_exit(0);
  }
  if(chat_mode == m_chattext)
  {
    if(*message != '\0')
    {
      if(specialcmd(chat, message) == 0)
      {
        int last_text_line = (chat->text_line + MAX_LINES - 1)%MAX_LINES;
        if (!(strcmp(chat->text[last_text_line].message, message) == 0 &&
            strcmp(chat->text[last_text_line].name, name) == 0)) {
          *(chat->text[chat->text_line].message) = '\0';
          chat->text[chat->text_line].atime = now;
          strncpy(chat->text[chat->text_line].name, name, NAME_LEN);
          strncpy(chat->text[chat->text_line].namecolor, namecolor, COLOR_LEN);
          strncpy(chat->text[chat->text_line].textcolor, textcolor, COLOR_LEN);
          strncpy(chat->text[chat->text_line].message, message, MSG_SIZE);
          if(CONV == 1)
            SET_FLAG(chat->text[chat->text_line].flags, TEXT_ITALIC);
          else
            RESET_FLAG(chat->text[chat->text_line].flags, TEXT_ITALIC);
          chat->text[chat->text_line].room = room;
          chat->text[chat->text_line].lang = LANG;
          chat->text_line = (++chat->text_line)%MAX_LINES;
          chat->user[userid].count ++;
#ifdef EXPORT_CHAT
#ifdef ALWAYS_EXPORT
          export_chat(chat, 0);
#endif /* ALWAYS_EXPORT */
#endif /* EXPORT_CHAT */
        }
      }
    }
  }
  if(isIMODE(ostype))
  {
    if(ilog > 0)
      chatlines = ilog;
    else
      chatlines = DEF_IMODE_MAXLINES;
  }
  else
  {
    if(slog > 0)
    {
      chatlines = slog;
      maxlines = slog;
    }
    else
      chatlines = NEWUSER_LINES + chat->user[userid].count;
  }
  /* get alert messages */
  if((int)chat->user[userid].memo[0] > 0 && alertmsg[0] == '\0')
  {
    memcpy(alertmsg, chat->user[userid].memo, MEMO_SIZE);
    chat->user[userid].memo[0] = '\0';
    for(i=1;i<MAX_USERS;i++)
    { /* clear other same name's alert */
      if((strcmp(chat->user[i].name, name) == 0) && 
         ((int)chat->user[i].memo[0] > 0))
      {
        chat->user[i].memo[0] = '\0';
      }
    }
  }
  memcpy(local, chat, sizeof(CHAT));

  unlock_chat(chat);
  release_chat(chat);

  if(chat_mode == m_logout)
  {
    header(local, 0);
    printf("<HR>\n");
    printf("<H3>%s</H3>\n", MSG(M_GOODBYE));
    printf("%s: ", MSG(M_LOGINTIME));
    if(isIMODE(ostype))
      printf("<BR>");
    printdate(NULL, logintime);
    printf("<BR>\n");
    printf("%s: ", MSG(M_LOGOUTTIME));
    if(isIMODE(ostype))
      printf("<BR>");
    printdate(NULL, now);
    if(isIMODE(ostype))
      printf("<HR>");
  }
  else
  {
    header(local, 1);
  }
  if(maxlines == 0) maxlines = DEF_MAX_LINES;
  if(chatlines > MAX_LINES) chatlines = MAX_LINES;
  if(chatlines > maxlines) chatlines = maxlines;
  printchat(local, chatlines);

  footer();
  free(local);
  return 0;
}

static char expiredate[256];
char* get_expire(tim)
int tim;
{
  char *wdays[] = {"Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Satuday", "Sunday"};
  char *mons[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
  time_t t = time(NULL) + tim;
  struct tm *gm;
  gm = gmtime(&t);
  sprintf(expiredate, "%s, %02d-%s-%04d %02d:%02d:%02d GMT",
    wdays[gm->tm_wday], gm->tm_mday, mons[gm->tm_mon],
    gm->tm_year+1900, gm->tm_hour, gm->tm_min, gm->tm_sec);
  return(expiredate);
}

void print_content()
{
  printf("Content-type: text/html; charset=%s\n", MSG(M_CHARSET));
  printf("Pragma: no-cache\n");
  printf("Cache-Control: no-cache\n");
  printf("Expires: Thu, 01 Dec 1994 16:00:00 GMT\n");
#ifdef COOKIE
  if(!isIMODE(ostype) && *name != '\0')
  {
    /* keep 30days */
    char *expdate = get_expire(60*60*24*30);
    if(chat_mode == m_entry)
    {
      int tmp_refresh = 0;
      if(*c_chatname == '\0')
        printf("Set-Cookie: chatname=%s; expires=%s\n", name, expdate);
      if(*c_autoclear != 0)
      {
        if((c_autoclear[0] - '0') != autoclear)
          printf("Set-Cookie: autoclear=%d; expires=%s\n", autoclear, expdate);
      }
      if(*c_refresh != 0)
      {
        tmp_refresh = atoi(c_refresh);
        if(tmp_refresh != refresh)
          printf("Set-Cookie: refresh=%d; expires=%s\n", refresh, expdate);
      }
    }
  }
#endif /* COOKIE */
  printf("\n");
  if(!isIMODE(ostype))
    printf("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n");
  printf("<HTML>\n");
}

void print_header(char *title, char *subtitle, int flg)
{
  printf("<HEAD>\n");
  if(!isIMODE(ostype))
  {
    if(oldmozilla)
    {
      if(LANG == get_LANG("ja") && OUTPUT_JCODE == OC_SJIS)
        printf("<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html;charset=x-sjis\">\n");
    }
    else
    {
      if(LANG == get_LANG("ja") && OUTPUT_JCODE == OC_EUC)
        printf("<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html;charset=ISO-2022-JP\">\n");
      else
        printf("<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html;charset=%s\">\n", MSG(M_CHARSET));
    }
  }
  if(title != 0)
  {
    printf("<TITLE>");
    nohtmlout(title);
    if(subtitle != 0)
    {
      printf("(");
      nohtmlout(subtitle);
      printf(")");
    }
    printf("</TITLE>\n");
  }
  if(flg)
    printf("</HEAD>\n");
}

void print_common(mode)
char *mode;
{
  char pname[256];
  char paddr[256];
  if(OUTPUT_JCODE == OC_SJIS)
  {
    strncpy(pname, name, sizeof(pname));
    strncpy(paddr, address, sizeof(paddr));
  }
  else
  {
    sjis2euc(pname, name);
    sjis2euc(paddr, address);
  }

  hidden_out("mode", mode);
  hidden_out("mlang", lang);
  hidden_out("jcode", jcode);
  hidden_out("name", pname);
  /*hidden_out("address", paddr);*/
  hidden_out("password", cpassword);
  hidden_out("bgcolor", bgcolor);
  hidden_out("txcolor", txcolor);
  if(imode)
    hidden_out("imode", "1");
  if(room != 0)
    hidden_out("room", "1");
}

void entry_form(chat)
CHAT *chat;
{
  header(chat, 0);
  if(autoreload)
  {
    printf("</BODY></HTML>\n");
    chat_exit(0);
  }
  printtopic(chat);
  if(!isIMODE(ostype))
  {
    printf("<HR>\n");
    printf("<H3>%s</H3>\n", MSG(M_USAGE));
  }
  printform("_self");
  if(isIMODE(ostype))
  {
    printf("<PRE>\n");
    printf("%s:\n", MSG(M_NAME));
    printf("<INPUT TYPE=TEXT NAME=name SIZE=15");
      printf(" MAXLENGTH=%d><BR>\n", NAME_LEN - 1);
    printf("%s:\n", MSG(M_ADDRESS));
    printf("<INPUT TYPE=TEXT NAME=address SIZE=15");
      printf(" MAXLENGTH=%d><BR>\n", ADDR_LEN - 1);
    printf("</PRE>\n");
  }
  else
  {
    char *selected = " SELECTED";
    int c_ref = 20;
    if(*c_refresh != '\0') c_ref = atoi(c_refresh);
    if(OUTPUT_JCODE == OC_XBM) c_ref = 0;
    printf("<TABLE>\n");
    printf("<TR><TD>%s</TD>\n", MSG(M_NAME));
    printf("<TD><INPUT TYPE=TEXT NAME=name SIZE=15");
      printf(" MAXLENGTH=%d><BR></TD></TR>\n", NAME_LEN - 1);
    printf("<TR><TD>%s</TD>", MSG(M_ADDRESS));
    printf("<TD><INPUT TYPE=TEXT NAME=address SIZE=40");
      printf(" MAXLENGTH=%d>\n", ADDR_LEN - 1);
    printf("<BR></TD></TR>\n");
    printf("</TABLE>\n");
    printf(" %s:<SELECT NAME=refresh>", MSG(M_RELOAD));
    if(OUTPUT_JCODE == OC_XBM) printf("<OPTION SELECTED>0");
    printf("<OPTION%s>5", c_ref == 5 ? selected : "");
    printf("<OPTION%s>10", c_ref == 10 ? selected : "");
    printf("<OPTION%s>15", c_ref == 15 ? selected : "");
    printf("<OPTION%s>20", c_ref == 20 ? selected : "");
    printf("<OPTION%s>30", c_ref == 30 ? selected : "");
    printf("<OPTION%s>50", c_ref == 50 ? selected : "");
    printf("<OPTION%s>80", c_ref == 80 ? selected : "");
    printf("</SELECT>\n");

    if(!isMAC(ostype))
    {
      if(oldmozilla)
      {
        printf("<INPUT TYPE=CHECKBOX NAME=autoclear");
        if(autoclear || (c_autoclear[0] == '1')) printf(" CHECKED");
        printf(">%s<BR>\n", MSG(M_AUTOCLEAR));
      }
      else
      {
        printf("<SCRIPT LANGUAGE=JavaScript>\n");
        printf("<!--\n");
        printf("document.write(\"<INPUT TYPE=CHECKBOX NAME=autoclear");
        if(autoclear || (c_autoclear[0] == '1')) printf(" CHECKED");
        printf(">%s\");\n", MSG(M_AUTOCLEAR));
        printf("// -->\n");
        printf("</SCRIPT>\n");
      }
      printf("<BR>\n");
    }
  }
  if(isIMODE(ostype))
  {
    hidden_out("mode", "entry");
    if(imode)
      hidden_out("imode", "1");
    if(noframe)
      hidden_out("noframe", "1");
  }
  else
  {
    hidden_out("mode", "entry");
    hidden_out("mlang", lang);
    hidden_out("jcode", jcode);
    hidden_out("bgcolor", bgcolor);
    hidden_out("txcolor", txcolor);
    hidden_out("mytitle", mytitle);
  }
  if(room != 0)
    hidden_out("room", "1");

  if(getenv("HTTP_REFERER") != NULL)
  {
    char tmp[ADDR_LEN], *p;
    strncpy(tmp, getenv("HTTP_REFERER"), ADDR_LEN);
    if(strncmp(tmp, "http:", 5) == 0)
    {
      /*if((p = (char *)strchr(tmp, '?')) != NULL) *p = '\0';*/
      hidden_out("referer", tmp);
    }
  }
  printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_JOIN));
  if(!isIMODE(ostype))
    printf("<INPUT TYPE=RESET VALUE=\"%s\">\n", MSG(M_CLEAR));
  printf("</FORM>\n");

  if(!isIMODE(ostype))
  {
    if(count_lang() > 1)
    {
      int i;
      printf("<HR>\n");
      printf("Language:");
      for(i = 0; i < count_lang(); i++)
      {
        if(i == LANG) 
        {
          printf(" [%s]", languages[i]);
        }
        else
        {
          printf(" <A HREF=\"%s?mlang=%s\">%s</A>", cgibin, languages[i], languages[i]);
        }
      }
    }
  
    printf("<HR>\n");
    printf("%s", MSG(M_CHATDESC));
    printlinks(chat);
  }
  footer();
  chat_exit(0);
}

void password_form()
{
  printf("Content-type: text/html; charset=%s\n\n", MSG(M_CHARSET));
  printf("<HTML>\n");
  print_header(MSG(M_PASSWORD), 0, 1);
  body_out(&body);
  printf("<FONT SIZE=4><STRONG>");
  htmlout(title, imode);
  printf("</STRONG></FONT>");
  printf("<HR>");

  printform("_self");
  printf("%s<BR>\n", MSG(M_PASSWORD));
  printf("<INPUT TYPE=PASSWORD NAME=pass SIZE=10>");
  if(isIMODE(ostype))
  {
    hidden_out("mode", "entry");
    hidden_out("name", name);
    hidden_out("address", address);
    if(imode)
      hidden_out("imode", "1");
    if(noframe)
      hidden_out("noframe", "1");
  }
  else
  {
    hidden_out("mode", "entry");
    hidden_out("name", name);
    hidden_out("address", address);
    printf("<INPUT TYPE=HIDDEN NAME=refresh VALUE=%d>", refresh);
    if(autoclear)
      hidden_out("autoclear", "on");
    hidden_out("mlang", lang);
    hidden_out("jcode", jcode);
    hidden_out("bgcolor", bgcolor);
    hidden_out("txcolor", txcolor);
    hidden_out("mytitle", mytitle);
  }
  if(room != 0)
    hidden_out("room", "1");
  printf("<BR><INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_JOIN));

  printf("</FORM>\n");
  footer();
  chat_exit(0);
}

void noframe_input()
{
  printf("<A NAME=top></A>");
  printform("_self");
  if(noframe)
    hidden_out("noframe", "1");
  print_common("chattext");
  if(isIMODE(ostype))
    printf("<INPUT TYPE=TEXT NAME=chat SIZE=16");
  else
    printf("<INPUT TYPE=TEXT NAME=chat SIZE=50");
  printf(" MAXLENGTH=1023><BR>\n");
  printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_MESSAGE));
  printf("<INPUT TYPE=RESET VALUE=\"%s\">\n", MSG(M_CLEAR));
  printf("</FORM>\n");
}

void enter_base()
{
  char pname[256];
  char ppass[256];
  char paddr[256];
  char ename[256];
  char eaddr[256];
  char pbgcolor[41];
  char ptxcolor[41];
  int inputsize = 20;
  if(OUTPUT_JCODE == OC_SJIS)
  {
    strncpy(ename, name, sizeof(ename));
    strncpy(eaddr, address, sizeof(eaddr));
  }
  else
  {
    sjis2euc(ename, name);
    sjis2euc(eaddr, address);
  }
  paramencode(pname, ename);
  paramencode(ppass, cpassword);
  paramencode(paddr, eaddr);
  paramencode(pbgcolor, bgcolor);
  paramencode(ptxcolor, txcolor);
  if(ismanager()) inputsize = 50;
  print_content();
  if(*mytitle != '\0')
    print_header(mytitle, 0, 1);
  else
    print_header(title, name, 1);
  printf("<FRAMESET ROWS=\"%d%%,*\">\n", inputsize);

  printf("<FRAME SRC=\"%s?", cgibin);
  printf("mode=input&name=%s&address=%s", pname, paddr);
  if(room == 1) printf("&room=1");
  printf("&password=%s", ppass);
  if(*bgcolor != '\0')
    printf("&bgcolor=%s", pbgcolor);
  if(*txcolor != '\0')
    printf("&txcolor=%s", ptxcolor);
  if(lang[0] != '\0')
    printf("&mlang=%s", lang);
  if(jcode[0] != '\0')
    printf("&jcode=%s", jcode);
  if(autoclear)
    printf("&autoclear=on");
  printf("\" NAME=chatin>\n");

  printf("<FRAME SRC=\"%s?", cgibin);
  printf("mode=chattext&name=%s&address=%s", pname, paddr);
  if(room == 1) printf("&room=1");
  printf("&password=%s", ppass);
  if(*bgcolor != '\0')
    printf("&bgcolor=%s", pbgcolor);
  if(*txcolor != '\0')
    printf("&txcolor=%s", ptxcolor);
  if(lang[0] != '\0')
    printf("&mlang=%s", lang);
  if(jcode[0] != '\0')
    printf("&jcode=%s", jcode);
  printf("\" NAME=chattext>\n");

  printf("</FRAMESET>\n");
  printf("<NOFRAMES>\n");
  body_out(&body);
  printf("<a href=\"%s?mode=chattext&noframe=1&name=%s&address=%s&password=%s&mlang=%s&jcode=%s\">",
    cgibin, pname, paddr, ppass, lang, jcode);
  htmlout(title, ostype);
  printf("</a>\n");
  printf("</BODY>\n");
  printf("</NOFRAMES>\n");
  printf("</HTML>\n");
  chat_exit(0);
}

void input_form(chat)
CHAT *chat;
{
  int i;

  print_content();
  print_header(title, 0, 0);

  if(autoclear)
  {
    printf("<SCRIPT LANGUAGE=JavaScript>\n");
    printf("<!--\n");
    printf("function erase(form) {\n");
    printf("  document.input2.chat.value = form.chat.value;\n");
    printf("  document.input2.submit();\n");
    printf("  if(form.chat.value != '') {\n");
    printf("    form.chat_bak.value = form.chat.value;\n");
    printf("  }\n");
    printf("  form.chat.value = '';\n");
    printf("  return false;\n");
    printf("}\n");
    printf("function undo(form) {\n");
    printf("  form.chat.value = form.chat_bak.value;\n");
    printf("}\n");
    printf("//-->\n");
    printf("</SCRIPT>\n");
  }
  printf("</HEAD>\n");

  body_out(&body);


  if(autoclear)
  {
    printf("<FORM NAME=input1 METHOD=%s ACTION=\"%s\" TARGET=chattext ONSUBMIT=\"return erase(this)\">\n",
      cgimtd, cgibin);
    hidden_out("chat_bak", " ");
    printf("<INPUT TYPE=TEXT NAME=chat SIZE=50");
      printf(" MAXLENGTH=1023 VALUE=\"\">\n");
  }
  else
  {
    printform("chattext");
    printf("<INPUT TYPE=TEXT NAME=chat SIZE=50");
      printf(" MAXLENGTH=1023 VALUE=\"\">\n");
  }
  if(room == 1)
    hidden_out("room", "1");
  hidden_out("mlang", languages[LANG]);
  print_common("chattext");
  printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_MESSAGE));
  printf("<INPUT TYPE=RESET NAME=reset VALUE=\"%s\">\n", MSG(M_CLEAR));
  if(autoclear)
    printf("<INPUT TYPE=BUTTON VALUE=\"%s\" ONCLICK=\"undo(this.form)\">\n", MSG(M_REEDIT));
  printf("</FORM><HR>\n");
  if(autoclear)
  {
    printf("<FORM NAME=input2 METHOD=%s ACTION=\"%s\" TARGET=chattext>\n",
      cgimtd, cgibin);
    hidden_out("chat", " ");
    if(room == 1)
      hidden_out("room", "1");
    print_common("chattext");
    printf("</FORM>\n");
  }

  printf("<TABLE><TR>\n");

  /* cancel */
  if(*password != '\0')
  {
    printf("<TD>\n");
    printform("chattext");
    printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", "cancel");
    printf("<INPUT TYPE=HIDDEN NAME=chat VALUE=\"/cancel\">\n");
    print_common("chattext");
    printf("</FORM>\n");
    printf("</TD>\n");
  }

  /* who */
  if(*password != '\0')
  {
    printf("<TD>\n");
    printform("chattext");
    printf("<INPUT TYPE=SUBMIT VALUE=\"who\">\n");
    printf("<INPUT TYPE=HIDDEN NAME=chat VALUE=\"/who\">\n");
    print_common("chattext");
    printf("</FORM>\n");
    printf("</TD>\n");
  }

  /* logout */
  printf("<TD>\n");
  printform("_top");
  printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_LEAVE));
  print_common("logout");
  printf("</FORM>\n");
  printf("</TD>\n");

  /* bbs */
  printf("<TD>\n");
  printform("_blank");
  printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_BBS));
  print_common("bbs");
  printf("</FORM>\n");
  printf("</TD>\n");

  printf("</TR></TABLE>\n");

  if(!ismanager())
  {
    printlinks(chat);
  }
  else
  { /* special costomizeable form */
    printf("<HR>\n");
    printform("_top");
    printf("%s:<INPUT TYPE=TEXT NAME=def_title", MSG(M_MODTITLE));
      printf(" VALUE=");
      print_string(def_title);
      printf(" SIZE=20><BR>\n");
    printf("%s:<INPUT TYPE=TEXT NAME=def_bgcolor", MSG(M_MODBGCOLOR));
      printf(" VALUE=\"%s\" SIZE=10>\n", def_bgcolor);
    printf("%s:<INPUT TYPE=TEXT NAME=def_txcolor", MSG(M_MODTXCOLOR));
      printf(" VALUE=\"%s\" SIZE=10><BR>\n", def_txcolor);
    printf("%s<BR>\n", MSG(M_MODLICOLOR));
    printf("<NOBR>LINK:<INPUT TYPE=TEXT NAME=def_licolor");
      printf(" VALUE=\"%s\" SIZE=10 MAXLENGTH=%d></NOBR>\n",
        def_licolor, COLOR_LEN - 1);
    printf("<NOBR>ALINK:<INPUT TYPE=TEXT NAME=def_alcolor");
      printf(" VALUE=\"%s\" SIZE=10 MAXLENGTH=%d></NOBR>\n",
        def_alcolor, COLOR_LEN - 1);
    printf("<NOBR>VLINK:<INPUT TYPE=TEXT NAME=def_vlcolor");
      printf(" VALUE=\"%s\" SIZE=10 MAXLENGTH=%d><NOBR><BR>\n",
        def_vlcolor, COLOR_LEN - 1);
    printf("%s:<INPUT TYPE=TEXT NAME=def_dncolor", MSG(M_MODDNCOLOR));
      printf(" VALUE=\"%s\" SIZE=10 MAXLENGTH=%d>\n",
        def_dncolor, COLOR_LEN - 1);
    printf("%s:<INPUT TYPE=TEXT NAME=def_dtcolor", MSG(M_MODDTCOLOR));
      printf(" VALUE=\"%s\" SIZE=10 MAXLENGTH=%d><BR>\n",
        def_dtcolor, COLOR_LEN - 1);
    printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_MODIFY));
    printf("<INPUT TYPE=RESET VALUE=\"%s\">\n", MSG(M_UNDO));
    print_common("customize");
    printf("</FORM>\n");

    printf("<HR>\n");
    printf("<H3>%s</H3>\n", MSG(M_LINK));

    printform("_top");
    for(i=0;i<MAX_LINKS;i++)
    {
      printf("URL:<INPUT TYPE=TEXT NAME=link_url");
        printf(" VALUE=\"%s\" SIZE=56 MAXLENGTH=%d><BR>\n",
          chat->links[i].url, ADDR_LEN - 1);
      printf(">>>>>><INPUT TYPE=TEXT NAME=link_label");
        printf(" VALUE=\"%s\" SIZE=32 MAXLENGTH=%d><BR>\n",
          chat->links[i].label_ja, LABEL_LEN - 1);
    }
    printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_MODIFY));
    printf("<INPUT TYPE=RESET VALUE=\"%s\">\n", MSG(M_UNDO));
    print_common("customize");
    printf("</FORM>\n");

    printf("<HR>\n");
    printform("_top");
    printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_INITIALIZE));
    print_common("reset");
    printf("</FORM>\n");
    printf("<HR>\n");
#ifdef EXPORT_INI
    if(access(EXPORT_INI, 0) == -1)
    {
      printf(MSG(M_CANNOTWRITE), EXPORT_INI);
      printf("<BR>\n");
    }
#endif
#ifdef EXPORT_USR
    if(access(EXPORT_USR, 0) == -1)
    {
      printf(iMSG(M_CANNOTWRITE), EXPORT_USR);
      printf("<BR>\n");
    }
#endif
#ifdef EXPORT_BBS
    if(access(EXPORT_BBS, 0) == -1)
    {
      printf(MSG(M_CANNOTWRITE), EXPORT_BBS);
      printf("<BR>\n");
    }
#endif
#ifdef EXPORT_CHAT
    if(access(EXPORT_CHAT, 0) == -1)
    {
      printf(MSG(M_CANNOTWRITE), EXPORT_CHAT);
      printf("<BR>\n");
    }
#endif
  }

  printf("</BODY>\n");
  printf("</HTML>\n");
  chat_exit(0);
}

void printlinks(chat)
CHAT *chat;
{
  int i;
  printf("<HR>\n");
  printf("<H3>%s</H3>\n", MSG(M_LINK));
  printf("%s", MSG(M_LINKS));
  printf("<UL>\n");
  for(i=0;i<MAX_LINKS;i++)
  {
    if(*chat->links[i].label_ja != '\0')
    {
      printf("<LI><A HREF=\"%s\" TARGET=_blank>", chat->links[i].url);
      htmlout(chat->links[i].label_ja, ostype);
      printf("</A><BR>\n");
    }
  }
  printf("</UL>\n");
}


void printdebug(chat)
CHAT *chat;
{
  int i, j;
  int users[MAX_USERS];

  /* sort by atime */
  for(i=0;i<MAX_USERS;i++) users[i] = i;
  for(i=0;i<MAX_USERS-1;i++)
  {
    for(j=i+1;j<MAX_USERS;j++)
    {
      int k;
      if(get_atime(chat, users[i]) < get_atime(chat, users[j]))
      {
        k = users[j];
        users[j] = users[i];
        users[i] = k;
      }
    }
  }

  printf("Content-type: text/plain\n\n");
  printf("memory size: %d version %2.2f\n", sizeof(CHAT), memversion/100.0);
  printf("HTTP_COOKIE: %s\n", cookie);
  for(i=0;i<MAX_USERS;i++)
  {
    int k;
    j = users[i];
    if(*chat->user[j].password != '\0')
      printf("*");
    else
      printf(" ");
    for(k = 0; k < MAX_PERMIT; k++)
    {
      if(chat->user[j].permit > k)
        printf("+");
      else
        printf(" ");
    }
    printf(" %3d ", users[i]);
    printf("["); printdate(NULL, chat->user[j].logintime); printf(" - ");
    printdate(NULL, chat->user[j].atime); printf("]");
    printf(" [%s:%d]", chat->user[j].name, chat->user[j].mid);
    printf(" [%s] [%s] [%d:%d]", chat->user[j].address, chat->user[j].agent,
        chat->user[j].login, chat->user[j].count);
    printf("[%s", IP2IPstr(chat->user[j].haddressip));
#ifdef HTTPD_HOSTNAMELOOKUP
    printf(":%s]", chat->user[j].haddress);
#else
    if(chat->user[j].haddress[0] == '\0')
    {
      strncpy(chat->user[j].haddress, gethostbyIP(chat->user[j].haddressip), HADDR_LEN);
    }
    printf(":%s]", chat->user[j].haddress);
#endif /* HTTPD_HOSTNAMELOOKUP */
    printf("[%s]", chat->user[j].referer);
    if(*chat->user[j].cookiename != '\0');
      printf(" (%s)", chat->user[j].cookiename);
    if(IS_FLAG(chat->user[j].flags, USER_SKIPTITLE))
      printf(" SKIPTITLE");
    if(IS_FLAG(chat->user[j].flags, USER_REFUSE))
      printf(" REFUSE");
    if(IS_FLAG(chat->user[j].flags, USER_TEXTHR))
      printf(" TEXTHR");
    if(IS_FLAG(chat->user[j].flags, USER_CRASH))
      printf(" CRASH");
    printf("\n");
  }
  printf("\n");
  for(i=0;i<MAX_DENYHOSTS;i++)
  {
    printf("%s", IP2IPstr(chat->denyhosts[i].address));
    printf(" ["); printdate(NULL, chat->denyhosts[i].atime); printf("]");
    if(*chat->denyhosts[i].name != '\0')
      printf(" [%s]", chat->denyhosts[i].name);
    if(chat->denyhosts[i].mid != 0)
      printf(" by %d", chat->denyhosts[i].mid);
    printf("\n");
  }
  for(i=0;i<MAX_LINKS;i++)
  {
    printf(" [%s][%s][%s]\n", chat->links[i].url, chat->links[i].label_ja,
      chat->links[i].label_en);
  }
}

void printdate(s, t)
char *s;
time_t t;
{
  struct tm *tim;
  time_t lt;
  char *mon[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
    "Sep", "Oct", "Nov", "Dec"};
  lt = t + SHIFT_TIME;
  tim = localtime(&lt);
  if(s != NULL)
  {
    if(strcmp(lang, "en") == 0)
    {
      sprintf(s, "%s %d, %02d:%02d%s",
        mon[tim->tm_mon], tim->tm_mday, tim->tm_hour, tim->tm_min, TZ_STRING);
    }
    else
    {
      sprintf(s, "%d/%d %02d:%02d%s",
        tim->tm_mon + 1, tim->tm_mday, tim->tm_hour, tim->tm_min, TZ_STRING);
    }
  }
  else
  {
    if(strcmp(lang, "en") == 0)
    {
      printf("%s %d, %02d:%02d%s",
        mon[tim->tm_mon], tim->tm_mday, tim->tm_hour, tim->tm_min, TZ_STRING);
    }
    else
    {
      printf("%d/%d %02d:%02d%s",
        tim->tm_mon + 1, tim->tm_mday, tim->tm_hour, tim->tm_min, TZ_STRING);
    }
  }
}

TEXT *gettextline(chat, lin)
CHAT *chat;
int lin;
{
  int i;
  i = (chat->text_line + MAX_LINES - lin - 1) % MAX_LINES;
  return(&chat->text[i]);
}

void printform(target)
char *target;
{
    if(isIMODE(ostype))
      target = 0; 
    printf("<FORM ACTION=\"%s\" ", cgibin);
    if((mozilla6 == 1) && (strcmp(target, "_blank") == 0))
      printf("METHOD=GET");
    else
      printf("METHOD=%s", cgimtd);
    if(target != 0)
      printf(" TARGET=%s", target);
    printf(">\n");
}

void printchat(chat, lines)
CHAT *chat;
int lines;
{
  int i, j, t = 0;
  int olin = 0;

  if(userreseted)
  {
    printf("<SCRIPT LANGUAGE=JavaScript>\n");
    printf("  top.location = \"%s\";\n", cgibin);
    printf("</SCRIPT>\n");
    return;
  }
  if (ilog > 0 && !isIMODE(ostype))
  {
    char pname[256];
    char ename[256];
    char pbgcolor[41];
    char ptxcolor[41];
    if(OUTPUT_JCODE == OC_SJIS)
      strncpy(ename, name, sizeof(ename));
    else
      sjis2euc(ename, name);
    paramencode(pname, ename);
    paramencode(pbgcolor, bgcolor);
    paramencode(ptxcolor, txcolor);

    printf("<SCRIPT LANGUAGE=JavaScript>\n");
    printf("  window.open('%s?auto=1&", cgibin);
    if(newtext) printf("newtext&");
    printf("mode=chattext");
    if(room != 0)
      printf("&room=%d", room);
    if(*name != '\0')
      printf("&name=%s", pname);
    if(*password != '\0')
      printf("&password=%s", cpassword);
    if(*bgcolor != '\0')
      printf("&bgcolor=%s", pbgcolor);
    if(*txcolor != '\0')
      printf("&txcolor=%s", ptxcolor);
    if(lang[0] != '\0')
      printf("&mlang=%s", lang);
    if(jcode[0] != '\0')
      printf("&jcode=%s", jcode);
    printf("&slog=%d", ilog);
    printf("','slog','');\n");
    printf("</SCRIPT>\n");
  }

  if((noframe || isIMODE(ostype)) && (chat_mode == m_chattext))
  {
    if(IS_FLAG(chat->user[userid].flags, USER_CRASH))
    {
      int i;
      for(i = 0; i < 10000; i++)
        printf("<BR>"); /* SEND CRASH DATA */
    }
    else
    {
      noframe_input();
    }
  }
  
  if((chat->topic.atime >= now - 60) && !isIMODE(ostype))
  {
    printtopic(chat);
    t = 1;
  }

  if(isIMODE(ostype))
    printf("<PRE>\n");
  else
    printf("<HR>\n");
  /* print chattext */
  for(i = j = 0; i < MAX_LINES; i++)
  {
    TEXT *txline = gettextline(chat, i);
    TEXT *nexttxline = gettextline(chat, i+1);
    if(txline->atime < nexttxline->atime) continue;
    if((txline->atime > 0) &&
       (*txline->message != '\0') &&
       (txline->room == room) )
    {
      if(isIMODE(ostype))
      {
        htmlout(txline->name, ostype);
        printf(">");
        htmlout(txline->message, ostype);
        if(idate)
        {
          printf("(");
          printdate(NULL, txline->atime);
          printf(")");
        }
        printf("\n");
        olin++;
      }
      else
      {
        int pline;
        pline = (chat->text_line + MAX_LINES - i - 1) % MAX_LINES;
        if(userproperty && (permission > 2))
          printf("<!--%d-->", pline);
        if(*txline->namecolor != '\0')
          printf("<FONT COLOR=\"%s\">", txline->namecolor);
        else
          printf("<FONT COLOR=\"%s\">", def_dncolor);
        if(IS_FLAG(txline->flags, TEXT_ITALIC)) printf("<I>");
        htmlout(txline->name, ostype);
        if(IS_FLAG(txline->flags, TEXT_ITALIC)) printf("</I>");
        printf("&gt; </FONT>");
        if(*txline->textcolor != '\0')
          printf("<FONT COLOR=\"%s\">", txline->textcolor);
        htmlout(txline->message, ostype);
        if(*txline->textcolor != '\0')
          printf("</FONT>");
        printf("  <FONT COLOR=\"%s\" SIZE=-1>(", def_dtcolor);
        printdate(NULL, txline->atime);
        printf(")</FONT>");
        if(deletebtn)
        {
          printform("chattext");
          printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_DELETE));
          print_common("chattext");
          printf("<INPUT TYPE=HIDDEN NAME=chat");
          printf(" VALUE=\"/delete %d\">\n", pline);
          printf("</FORM>\n");
        }
        printf("<BR>\n");

#ifdef TEXT_HR
        printf("<HR>");
#else
        if(texthr)
          printf("<HR>");
#endif
        olin++;
      }
      if(++j >= lines) break;
      if(newtext && (txline->atime < (lastatime - 60)) && (ilog == 0)) break;
    }
  }
  if(isIMODE(ostype))
  {
    printdate(NULL, now);
    if(olin > 5)
      printf(" <A HREF=#top>^</A>");
    printf("\n");
    /*show_users(chat, 0);*/
    printf("</PRE>\n");
  }
  else
  {
    if(t == 0)
      printtopic(chat);
  }
  if((noframe || isIMODE(ostype)) && chat_mode == m_chattext)
  {
    printf("<HR>");
    if(ostype != OS_JPHONE)
    {
      printf("<TABLE>\n");
      printf("<TD>\n");
    }
    /* logout */
    printform("_top");
    printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_LEAVE));
    print_common("logout");
    printf("</FORM>\n");
    if(ostype != OS_JPHONE)
    {
      printf("</TD>\n");
      printf("<TD>\n");
    }
    /* bbs */
    printform("_top");
    printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_BBS));
    print_common("bbs");
    printf("</FORM>\n");
    if(ostype != OS_JPHONE)
    {
      printf("</TD>\n");
      printf("</TABLE>\n");
    }
  }

}

void printtopic(chat)
CHAT *chat;
{
  if(chat != NULL)
  {
    /* print topic */
    if((chat->topic.atime > 0) &&
       (*chat->topic.message != '\0'))
    {
      printf("<HR>\n");
      printf("%s: ", MSG(M_TOPIC));
      htmlout(chat->topic.message, ostype);
      printf(" (");
      printdate(NULL, chat->topic.atime);
      printf(" "); htmlout(chat->topic.name, ostype); printf(")\n");
    }
  }
}

void printbbs(chat, line)
CHAT *chat;
int line;
{
  int i, j = 0, k = 0;
  char pname[256];
  char paddr[256];
  int bbslines = DEF_BBSLINES;

  if(isIMODE(ostype)) bbslines = DEF_IMODE_BBSLINES;
  if(OUTPUT_JCODE == OC_SJIS)
  {
    strncpy(pname, name, sizeof(pname));
    strncpy(paddr, address, sizeof(paddr));
  }
  else
  {
    sjis2euc(pname, name);
    sjis2euc(paddr, address);
  }
  print_content();
  print_header(title, MSG(M_BBS), 1);
  body_out(&body);
  if(!isIMODE(ostype))
  {
    printf("<H3>"); htmlout(title, ostype);
    printf("(%s)</H3>\n", MSG(M_BBS));
    printf("%s: ", MSG(M_NAME)); htmlout(name, ostype); printf("<BR>\n");
  }

  if(permission > 0)
  {
    printform("_self");
    if(isIMODE(ostype))
      printf("<TEXTAREA NAME=chat ROWS=2 COLS=16>");
    else
      printf("<TEXTAREA NAME=chat ROWS=3 COLS=60>");
    printf("%s", message);
    printf("</TEXTAREA><BR>\n");
    printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_BBSWRITE));
    printf("<INPUT TYPE=RESET VALUE=\"%s\">\n", MSG(M_CLEAR));
    printf("<INPUT TYPE=CHECKBOX NAME=secure>%s\n", MSG(M_SECURE));
    print_common("bbs");
    printf("</FORM>\n");
  }

  if (line > 0)
    printf("<HR>\n");
  /* print bbs */
  for(i = line; i < MAX_BBS; i++)
  {
    int pline;
    pline = (chat->bbs_line - 1 - i + MAX_BBS)%MAX_BBS;
    if((chat->bbs[pline].atime > 0) &&
       (*chat->bbs[pline].message != '\0') &&
       (j <= bbslines) )
    {
      int sec = 0;
      if(IS_FLAG(chat->bbs[pline].flags, SECURE_BBS))
        sec = 1;
      if(permission == 0 && sec == 1) continue;
      if(j == bbslines)
      {
        k = i;
        break;
      }
      printf("<HR>\n");
      if(!isIMODE(ostype))
      {
        printf("<A NAME=%d>", pline);
        printf("<FONT SIZE=4 COLOR=\"%s\">", def_dncolor);
      }
      if((userid >= 0) && (*password != '\0') && (chat->bbs[pline].atime > lastbtime))
        printf("<B>[New]</B> ");
      if(sec)
        printf("*** ");
      printdate(NULL, chat->bbs[pline].atime);
      if(isIMODE(ostype))
      {
        printf("<BR>");
        htmlout(chat->bbs[pline].name, ostype);
        printf(":");
      }
      else
      {
        printf(" (");
        htmlout(chat->bbs[pline].name, ostype);
        printf(")");
      }
      if(sec)
        printf(" ***");
      if(!isIMODE(ostype))
      {
        printf("</A>");
        printf("</FONT><BR>");
      }
      printf("\n<PRE>\n");
      htmlout(chat->bbs[pline].message, ostype);
      printf("\n</PRE>\n");
      if(strcmp(name, chat->bbs[pline].name) == 0 || ((permission > 2) && !isIMODE(ostype)))
      {
        printform("_self");
        printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", MSG(M_DELETE));
        print_common("bbs");
        printf("<INPUT TYPE=HIDDEN NAME=chat");
        printf(" VALUE=\"/delete %d\">\n", pline);
        printf("</FORM>\n");
      }
      j++;
    }
  }
  if (k > 0)
  {
    printf("<HR>\n");
    printform("_self");
    printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", "Old Topics >>");
    print_common("bbs");
    printf("<INPUT TYPE=HIDDEN NAME=bbsline");
    printf(" VALUE=\"%d\">\n", k);
    printf("</FORM>\n");
  }
  if(isIMODE(ostype))
  {
    printform("_self");
    printf("<INPUT TYPE=SUBMIT VALUE=\"%s\">\n", "Chat >>");
    if(noframe)
      hidden_out("noframe", "1");
    print_common("chattext");
    printf("<INPUT TYPE=HIDDEN NAME=bbsline");
    printf(" VALUE=\"%d\">\n", k);
    printf("</FORM>\n");
  }
  footer();
}

#ifndef WIN32
void sigalm()
{
  signal(SIGALRM, SIG_IGN);
  if(outhead == 0)
    header(NULL, 0);
  printf("Timeout!\n");
  if(chat != NULL)
  {
    unlock_chat(chat);
    release_chat(chat);
  }
  footer();
  exit(1);
}
#endif /* !WIN32 */

void sigterm(sig)
int sig;
{
  signal(SIGSEGV, SIG_IGN);
  signal(SIGTERM, SIG_IGN);
  signal(SIGINT, SIG_IGN);
  if(outhead == 0)
    header(NULL, 0);
  printf("Terminated!!(%d)\n", sig);
  if(chat != NULL)
  {
    unlock_chat(chat);
    release_chat(chat);
  }
  footer();
  exit(1);
}

void killtag(str)
char *str;
{
  int i;
  for(i=0;str[i] != '\0';i++)
    if(str[i] == '<' || str[i] == '>') str[i] = ' ';
  for(i=strlen(str)-1;i>=0;i--)
    if(str[i] > 0 && str[i] <= ' ') str[i] = '\0';
}

char *getcgiarg()
{
  int param_len;
  int postmethod = 0;
  char *param;
  static char *null = "";


  if(getenv("REQUEST_METHOD") != NULL)
  {
    if(strcmp((char *)getenv("REQUEST_METHOD"), "POST") == 0)
      postmethod = 1;
  }
  else
    return(null);

  if(postmethod == 0)
  {
    char *p;
    if((p = getenv("QUERY_STRING")) != NULL)
    {
      if((param = malloc(strlen(p)+1)) != NULL)
        strcpy(param, p);
      else
        param = null;
    }
    else
      param = null;
  }
  else
  {
    if(getenv("CONTENT_LENGTH") != NULL)
    {
      param_len = atoi(getenv("CONTENT_LENGTH"));
      if((param = malloc(param_len+1)) != NULL)
        fgets(param, param_len+1, stdin);
      else
        param = null;
    }
    else
      param = null;
  }
  return(param);
}

int get_userentry()
{
  int i;
  
  /* check manager entry */
  if(strcmp(name, managerid) == 0)
  {
    if(*password == '\0')
      return(-2);
    if(strcmp(password, managerpass) == 0)
      return(0);
    *name = '\0';
    return(-1);
  }
  /* check user entry */
  if(*password != '\0') /* with a password */
  {
    if(strcmp(password, managerpass) == 0)
    {
      if(strcmp(name, managerid) == 0)
        return(0);
      *name = '\0';
      return(-1);
    }
    /* search same handle name */
    for(i=1;i<MAX_USERS;i++)
    {
      if(strcmp(chat->user[i].name, name) == 0)
      {
        if(*chat->user[i].password == '\0')
        {
          if(*password != '\0')
            export_user(EXPORT_USER, &chat->user[i], 0);
          return(i);
        }
        if(strcmp(chat->user[i].password, password) == 0)
          return(i);
        /* invalid password */
        *name = '\0';
        return(-1);
      }
    }
    /* search same password */
    for(i=1;i<MAX_USERS;i++)
    {
      if(strcmp(chat->user[i].password, password) == 0)
      {
        if(strcmp(chat->user[i].name, name) == 0)
          return(i);
        if(compareIP(chat->user[i].haddressip, haddressip, 0xffffff00) == 0)
        {
          /* change handle name */
          export_user(EXPORT_USER, &chat->user[i], name);
          return(i);
        }
        *name = '\0';
        return(-1);
      }
    }
  }
  else /* without password */
  {
    for(i=1;i<MAX_USERS;i++)
    {
      if(strcmp(chat->user[i].name, name) == 0)
      {
        if(*chat->user[i].password != '\0')
        {
          /* need password */
          return(-2);
        }
        if(compareIP(chat->user[i].haddressip, haddressip, 0xffffff00) == 0)
          return(i);
      }
    }
  }

  return(-1);
}

int new_userentry()
{
  int i, j = 0;
  int older;

  /* search null entry */
  for(i=1;i<MAX_USERS;i++)
  {
    if((chat->user[i].atime == 0) && (chat->user[i].permit == 0) &&
       (*chat->user[i].password == '\0'))
      return(i);
    if(chat->user[i].atime != 0) j++;
  }
  for(i=1;i<MAX_USERS;i++)
  {
    if((chat->user[i].atime == 0) && (chat->user[i].permit == 0))
      return(i);
  }
  if(j > MAX_USERS/2)
  {
    for(i=1;i<MAX_USERS;i++)
    {
      if((chat->user[i].atime == 0) && (chat->user[i].permit < 2))
        return(i);
    }
  }
  older = 1;
  for(i=2;i<MAX_USERS;i++)
  {
    if((chat->user[older].atime > chat->user[i].atime) &&
       (chat->user[i].permit <= DEFAULT_PERMIT) &&
       (*chat->user[i].password == '\0'))
      older = i;
  }
  if(chat->user[older].atime > now - 3600)
  {
    for(i=2;i<MAX_USERS;i++)
    {
      if((chat->user[older].atime > chat->user[i].atime) &&
         (chat->user[i].permit <= DEFAULT_PERMIT))
        older = i;
    }
  }
  return(older);
}

void clearuserinfo(user)
USERINFO *user;
{
  memset((void *)user, 0, sizeof(USERINFO));
  user->maxlines = DEF_MAX_LINES;
  user->permit = DEFAULT_PERMIT;
}

int setuserinfo(chat)
CHAT *chat;
{
  int i;
  int newuser = 0;

  if(userid != -1) return userid;

  i = get_userentry();
  if(i == -1)
  { /* new user */
    USERINFO tmpuser;
    if(get_lockmode(chat) >= CHAT_LOCK_REGISTED)
    {
      scerr(MSG(M_CANNOTUSE));
    }
    if(*name == '\0')
    {
      sendmessage(alertmsg, sizeof(alertmsg), MSG(M_INVALIDPASS), MT_ERROR);
      back = 1;
      scerr(MSG(M_INVALIDPASS));
    }
    if((i = new_userentry()) == -1)
    {
      scerr(MSG(M_CANNOTUSE));
    }
#ifdef EXPORT_USER
    memset(&tmpuser, 0, sizeof(USERINFO));
    if(read_user(EXPORT_USER, &tmpuser, name, password) != 0)
    {
      memcpy(&chat->user[i], &tmpuser, sizeof(USERINFO));
      if(strcmp(chat->user[i].password, password) != 0)
      {
        if(*password == '\0')
        {
          unlock_chat(chat);
          release_chat(chat);
          password_form();
        }
        sendmessage(alertmsg, sizeof(alertmsg), MSG(M_INVALIDPASS), MT_ERROR);
        back = 1;
        scerr(MSG(M_INVALIDPASS));
      }
    }
    else
#endif /* EXPORT_USER */
      clearuserinfo(&chat->user[i]);
    newuser = 1;
  }
  else if(i == -2)
  {
    unlock_chat(chat);
    release_chat(chat);
    password_form();
  }
  strncpy(chat->user[i].name, name, NAME_LEN);
  strncpy(chat->user[i].password, password, PASS_LEN);
  strncpy(chat->user[i].haddress, haddress, HADDR_LEN);
  strncpy(chat->user[i].agent, agent, AGENT_LEN);
  chat->user[i].haddressip = haddressip;
  if(autoreload == 0 || chat->user[i].atime < now - AUTOLOGOUT*2)
    chat->user[i].ctime = now;
  lastatime = chat->user[i].atime;
  loginstat = chat->user[i].login;
  if (chat_mode != m_bbs)
  {
    chat->user[i].atime = now;
    chat->user[i].room = room;
  }
  switch(chat_mode)
  {
  case m_logout:
    chat->user[i].login = 0;
    break;
  case m_entry:
    {
      chat->user[i].logintime = now;
      strncpy(chat->user[i].address, address, ADDR_LEN);
      strncpy(chat->user[i].referer, referer, ADDR_LEN);
      if(CONV == 1)
        SET_FLAG(chat->user[i].flags, USER_ITALIC);
      else
        RESET_FLAG(chat->user[i].flags, USER_ITALIC);
      if(!isIMODE(ostype))
      {
        chat->user[i].refresh = refresh;
      }
      else
      {
        lastatime = 0;
      }
#ifdef COOKIE
      if(*c_chatname != '\0')
      {
        strncpy(chat->user[i].cookiename, c_chatname, NAME_LEN);
      }
      else
        *chat->user[i].cookiename = '\0';
#endif /* COOKIE */
    }
  default:
    if(noframe || isIMODE(ostype))
    {
      if(*address != '\0')
        strncpy(chat->user[i].address, address, ADDR_LEN);
      if (chat->user[i].logintime == 0 || chat->user[i].login == 0)
        chat->user[i].logintime = now;
    }
    chat->user[i].login = 1;
    maxlines = chat->user[i].maxlines;
    skiptitle = IS_FLAG(chat->user[i].flags, USER_SKIPTITLE);
    texthr = IS_FLAG(chat->user[i].flags, USER_TEXTHR);
    if(*chat->user[i].namecolor != '\0')
      strncpy(namecolor, chat->user[i].namecolor, COLOR_LEN);
    if(*chat->user[i].textcolor != '\0')
      strncpy(textcolor, chat->user[i].textcolor, COLOR_LEN);
  }
  if(ismanager())
    chat->user[i].permit = MAX_PERMIT;
  else if(newuser && *password != '\0')
    export_user(EXPORT_USER, &chat->user[i], 0);

  if(get_lockmode(chat) >= CHAT_LOCK_PERMIT && chat->user[i].permit < 1)
  {
    scerr(MSG(M_CANNOTUSE));
  }

  return(i);
}

char *getparam(dd, maxlen, src)
char *dd;
int maxlen;
char *src;
{
  char d[MSG_SIZE], *dst = d, *p;
  int i = 0, jis = 0;
  int j, k, c;

  while(1)
  {
    if(*src == '&' || *src == 0x0d || *src == 0x0a || *src == 0) break;
    if(*src == '%')
    {
      int c;
      c = (h2i(*(src+1))<<4) | h2i(*(src+2));
      if((c >= 0 && c < 0x20) && (c != 0x0d) && (c != 0x0a) && (c != 0x1b) && (c != 0x0f)) c = 0x20; /* invalid control chars */
      dst[i++] = c;
      src+=3;
    }
    else if(*src == '+')
    {
      dst[i++] = ' ';
      src++;
    }
    else
      dst[i++] = *src++;
    if(i >= maxlen - 1)
    {
       while((*src != '&') && (*src != 0x0d) &&
             (*src != 0x0a) && (*src != 0)) src++;
       break;
    }
  }
  dst[i] = '\0';
  while(*src == 0x0d || *src == 0x0a) src++;
  if(*src == '&') src++;
  if(!isIMODE(ostype) && ((p = (char *)strchr(d, ESC)) != NULL))
  {
    if( (strlen(p) > 3)
       && (*(p+1) == '$')
       && (*(p+2) == 'B' || *(p+2) == '@')) jis = 1;
  }
  
  if((OUTPUT_JCODE == OC_SJIS) && (jis == 0))
    strncpy(dd, d, maxlen);
  else
    euc2sjis(dd, d);
#ifdef IMODE
  k = strlen(dd);
  for(i = 0, j = 0; i < k; i++)
  {
    if(sjis1st(dd[i]))
    {
        dd[j++] = dd[i++];
        dd[j++] = dd[i];
    }
    else
    {
      if((i <= k - 7) && (dd[i] == '&') && (dd[i+1] == '#'))
      {
        c = (dd[i+2] - '0')*10000 + (dd[i+3] - '0')*1000 + (dd[i+4] - '0')*100 + (dd[i+5] - '0')*10 + (dd[i+6] - '0');
        if (((c >= 63808) && (c <= 63920)) || ((c >= 63552) && (c <= 63567)) || ((c >= 63646) && (c <= 63716)) || ((c >= 63731) && (c <= 63743)))
        {
          dd[j++] = (unsigned char)((c >> 8) & 0xff);
          dd[j++] = (unsigned char)(c & 0xff);
          i += 6;
          if(dd[i+1] == ';') i++;
        }
        else
        {
          dd[j++] = dd[i];
        }
      }
      else
      {
        dd[j++] = dd[i];
      }
    }
  }
  dd[j] = '\0';
#endif /* IMODE */
  return(src);
}

void paramencode(dst, src)
char *dst;
char *src;
{
  int i;
  for(i=0;src[i] != '\0';i++)
  {
    if(src[i] == '%' || src[i] == '+' || src[i] == '#' || src[i] == '"' ||
       src[i] == '&' || src[i] == '\'' || src[i] == '\\' ||
       (src[i]>=0 && src[i]<=' '))
    {
      *dst++ = '%';
      *dst++ = i2h((src[i]>>4)&0x0f);
      *dst++ = i2h(src[i]&0x0f);
    }
    else if((src[i]&0x80) != 0) /* kanji */
    {
      *dst++ = '%';
      *dst++ = i2h((src[i]>>4)&0x0f);
      *dst++ = i2h(src[i]&0x0f);
      i++;
      *dst++ = '%';
      *dst++ = i2h((src[i]>>4)&0x0f);
      *dst++ = i2h(src[i]&0x0f);
    }
    else
      *dst++ = src[i];
  }
  *dst = '\0';
}

time_t get_atime(chat, user)
CHAT *chat;
int user;
{
  if(chat->user[user].atime > chat->user[user].btime)
    return chat->user[user].atime;
  else
    return chat->user[user].btime;
}

void show_users(chat, showaddr)
CHAT *chat;
int showaddr;
{
  int i, j;
  int u, users[MAX_USERS];

  /* get active users */
  for(u=0,i=0;i<MAX_USERS;i++)
  {
    if((get_atime(chat, i) > now - AUTOLOGOUT*2) &&
        (*chat->user[i].name != '\0'))
    {
      if((chat->user[i].login == 0) && (get_atime(chat, i) < now - 30)) continue;
      if((chat->user[i].refresh > 0) && (get_atime(chat, i) < now - AUTOLOGOUT)) continue;
      users[u++] = i;
    }
  }

  /* sort by count */
  for(i=0;i<u-1;i++)
  {
    for(j=i+1;j<u;j++)
    {
      int k;
      if(chat->user[users[i]].count < chat->user[users[j]].count)
      {
        k = users[j];
        users[j] = users[i];
        users[i] = k;
      }
    }
  }

#ifdef IMODE
  if(isIMODE(ostype))
  {
    for(j=0;j<u;j++)
    {
      int i;
      i = users[j];
      printf("[");
      htmlout(chat->user[i].name, ostype);
      if(chat->user[i].login == 0)
        printf("(bye!)");
      else if(chat->user[i].login == 2)
        printf("(dead!)");
      if(chat->user[i].ctime < now - SLEEPTIME && chat->user[i].login == 1)
      {
        int h;
        h = (now - chat->user[i].ctime)/3600;
        printf("(Zzz");
        if(h > 0)
          printf("%d", h);
        printf(")");
      }
      printf("]");
    }
    return;
  }
#endif /* IMODE */

  for(j=0;j<u;j++)
  {
    int i;
    int blink = 0;
    int hlink = 0;
    int smalluser = 0;
    i = users[j];

    if(userproperty)
      printf("<BR>\n<NOBR>");

    if(room != chat->user[i].room)
    {
      smalluser = 1;
      printf("<FONT SIZE=1>");
    }
    printf(" [");
    if((get_atime(chat, i) < now - BLINKTIME) || (chat->user[i].login != 1))
    {
      printf("<BLINK>");
      blink = 1;
    }
    if((*chat->user[i].address != '\0') && (showaddr == 1) &&
        (chat->user[i].login == 1))
    {
      if(strchr(chat->user[i].address, ':') != NULL)
        printf("<A HREF=\"%s?URL=%s\" TARGET=\"_blank\">", cgibin, chat->user[i].address);
      else
        printf("<A HREF=\"mailto:%s\" TARGET=_top>", chat->user[i].address);
      if(IS_FLAG(chat->user[i].flags, USER_ITALIC)) printf("<I>");
      if(permission >= 2 && chat->user[i].altime > now - BOLDTIME) printf("<B>");
      htmlout(chat->user[i].name, ostype);
      if(permission >= 2 && chat->user[i].altime > now - BOLDTIME) printf("</B>");
      if(IS_FLAG(chat->user[i].flags, USER_ITALIC)) printf("</I>");
      hlink = 1;
    }
    else
    {
      if(IS_FLAG(chat->user[i].flags, USER_ITALIC)) printf("<I>");
      if(permission >= 2 && chat->user[i].altime > now - BOLDTIME) printf("<B>");
      htmlout(chat->user[i].name, ostype);
      if(permission >= 2 && chat->user[i].altime > now - BOLDTIME) printf("</B>");
      if(IS_FLAG(chat->user[i].flags, USER_ITALIC)) printf("</I>");
    }
    if(chat->user[i].login == 0)
      printf("(bye!)");
    else if(chat->user[i].login == 2)
      printf("(dead!)");
    if(chat->user[i].ctime < now - SLEEPTIME && chat->user[i].login == 1)
    {
      int h;
      h = (now - chat->user[i].ctime)/3600;
      printf("(Zzz");
      if(h > 0)
        printf("%d", h);
      printf(")");
    }
    if(hlink) printf("</A>");
    if(blink) printf("</BLINK>");
    printf("]");
    if(smalluser) printf("</FONT>");

    if(userproperty)
    {
      int ref = 0;
      if(*chat->user[i].password != '\0')
        printf("*");
      if(chat->user[i].permit)
        printf("+");
      if(permission >= 2)
      {
        int j;
        for(j = 2; j <= chat->user[i].permit; j++)
        {
          printf("+");
          if(j == permission) break;
        }
      }
      printf(":&lt;%d&gt;:", chat->user[i].count);
      if(*chat->user[i].referer != '\0')
      {
        printf("<A HREF=\"%s?URL=%s\" TARGET=\"_blank\">", cgibin, chat->user[i].referer);
        ref = 1;
      }
      printf("%s", IP2IPstr(chat->user[i].haddressip));
      if(ref) printf("</A>");
#ifdef HTTPD_HOSTNAMELOOKUP
      printf("(%s):", chat->user[i].haddress);
#else
      if(chat->user[i].haddress[0] == '\0')
      {
        strncpy(chat->user[i].haddress, gethostbyIP(chat->user[i].haddressip), HADDR_LEN);
      }
      printf("(%s):", chat->user[i].haddress);
#endif /* HTTPD_HOSTNAMELOOKUP */
      printf("%s", chat->user[i].agent);
#ifdef COOKIE
      if((permission >= 1) && (*chat->user[i].cookiename != '\0'))
      {
        printf(" (%s)", chat->user[i].cookiename);
      }
#endif /* COOKIE */
      printf("</NOBR>");
    }

  }
  printf("\n");
}


void header(chat, showaddr)
CHAT *chat;
int showaddr;
{
  int refresh = 0;
  int urcnt = 0;
  if(userid != -1 && chat_mode == m_chattext) refresh = chat->user[userid].refresh;
  if(noframe || isIMODE(ostype) || slog > 0)
  {
     refresh = 0;
  }
  print_content();
  print_header(title, 0, 0);
  if(refresh > 0)
  {
    char pname[256];
    char ppass[256];
    char ename[256];
    char pbgcolor[41];
    char ptxcolor[41];
    if(OUTPUT_JCODE == OC_SJIS)
      strncpy(ename, name, sizeof(ename));
    else
      sjis2euc(ename, name);
    paramencode(pname, ename);
    paramencode(ppass, cpassword);
    paramencode(pbgcolor, bgcolor);
    paramencode(ptxcolor, txcolor);
    printf("<META HTTP-EQUIV=\"Refresh\" CONTENT=\"%d;", refresh);
    printf("URL=%s?auto=1&", cgibin);
    if(newtext) printf("newtext&");
    printf("mode=chattext");
    if(room != 0)
      printf("&room=%d", room);
    if(*name != '\0')
      printf("&name=%s", pname);
    if(*password != '\0')
      printf("&password=%s", ppass);
    if(*bgcolor != '\0')
      printf("&bgcolor=%s", pbgcolor);
    if(*txcolor != '\0')
      printf("&txcolor=%s", ptxcolor);
    if(lang[0] != '\0')
      printf("&mlang=%s", lang);
    if(jcode[0] != '\0')
      printf("&jcode=%s", jcode);
    printf("\">\n");
  }
  printf("</HEAD>\n");
  if(isIMODE(ostype))
  {
    printf("<BODY>\n");
    alertmessages();
  }
  else
  {
    alertmessages();
    body_out(&body);
  }
  if(hasalert && !isIMODE(ostype))
  {
    printf("<SCRIPT LANGUAGE=JavaScript>\n");
    printf("showalert()\n");
    printf("</SCRIPT>\n");
    if(alertbtn)
    {
      printf("<FORM><INPUT TYPE=button VALUE=\"alert!\" ONCLICK=\"showalert()\">\n");
      printf("</FORM>\n");
    }
  }
  if(back) historyback();
  /* count unread bbs messages */
  if(userid >= 0 && *password != '\0')
  { /* count unread bbs */
    int i;
    for(i=0;i<MAX_BBS;i++)
    {
      int lin = (chat->bbs_line-1-i+MAX_BBS)%MAX_BBS;
      if(chat->bbs[lin].atime == 0) continue;
      if( (chat->user[userid].btime < chat->bbs[lin].atime) &&
          !(IS_FLAG(chat->bbs[lin].flags, SECURE_BBS) && (permission == 0)) )
        urcnt ++;
    }
  }
  if(isIMODE(ostype))
  {
    if(!(chat_mode == m_chattext || chat_mode == m_bbs))
    {
      printf("%s<BR>\n", title);
    }
    if(urcnt > 0)
      printf("%s%s=%d<BR>\n", MSG(M_BBS), MSG(M_UNREADBBS), urcnt);
  }
  else if(!skiptitle)
  {
    if(chat == NULL || *chat->titleimage == '\0')
    {
      if(chat_mode == m_logout)
        printf("<A HREF=\"%s\">", cgibin);
      printf("<FONT SIZE=4><STRONG>");
      htmlout(title, ostype);
      printf("</STRONG></FONT>");
      if(chat_mode == m_logout)
        printf("</A>");
    }
    else
    {
      printf("<IMG SRC=\"%s\" BORDER=0 ALIGN=top ALT=\"%s\">", chat->titleimage, title);
    }
    if(refresh > 0 && showaddr)
    {
      printf("<FONT COLOR=\"%s\"> (refresh=%dsec.", def_dtcolor, refresh);
      if(room != 0) printf("/room=%d", room);
      if(get_lockmode(chat) > 0) printf("/lock=%d", get_lockmode(chat));
      if(urcnt > 0)
        printf("/%s %s=%d\n", MSG(M_BBS), MSG(M_UNREADBBS), urcnt);
      printf(")</FONT>\n");
    }
  }
  outhead = 1;
  if(chat == NULL) return;
  if(isIMODE(ostype) && (userproperty == 0) && (chat_mode == m_chattext)) return;
  printf("<P>\n");
  printf("%s: ", MSG(M_MEMBERS));
  show_users(chat, showaddr);

}

void footer()
{
  if(isIMODE(ostype))
  {
    if(!(chat_mode == m_chattext || chat_mode == m_bbs))
    {
      printf("<A HREF=\"%s\">", official_url_imode);
      printf("%s ver%2.2f", software_name, (double)version/100);
      if(patchlevel > 0) printf("%c", '\x60' + (char)patchlevel);
      if(patchlevel < 0) printf("beta%d", -patchlevel);
      printf("</A>\n");
    }
  }
  else
  {
    printf("<HR>\n");
    printf("<P ALIGN=right>");
    printf("<A HREF=\"%s\" TARGET=\"_blank\">", official_url);
    printf("%s ver%2.2f", software_name, (double)version/100);
    if(patchlevel > 0) printf("%c", '\x60' + (char)patchlevel);
    if(patchlevel < 0) printf("beta%d", -patchlevel);
    printf("</A></P>\n");
  }
  printf("</BODY>\n");
  printf("</HTML>\n");
}

void scerr(msg)
char *msg;
{
  header(NULL, 0);
  printf("<H1>%s</H1>\n", msg);
  footer();
  if(chat != NULL)
  {
    unlock_chat(chat);
    release_chat(chat);
  }
  exit(1);
}

void event_form(chat)
CHAT *chat;
{
  int i;

  print_content();
  print_header(title, 0, 0);
  printf("</HEAD>\n");

  body_out(&body);

  printf("</BODY>\n");
  printf("</HTML>\n");
  chat_exit(0);
}

void print_forward(url)
char *url;
{
  print_content();
  print_header(title, 0, 0);
  if(!isIMODE(ostype)) {
    printf("<META HTTP-EQUIV=\"Refresh\" CONTENT=\"0;URL=%s\">\n", url);
  }
  printf("</HEAD>\n");
  if(!isIMODE(ostype)) {
    printf("<SCRIPT LANGUAGE=\"JavaScript\">\n");
    printf("<!--\n");
    printf("document.location='%s';\n", url);
    printf("-->\n");
    printf("</SCRIPT>\n");
  }
  printf("<BODY>\n");
  printf("<A HREF=\"%s\">%s</a>\n", url, url);
  printf("</BODY>\n");
  printf("</HTML>\n");
}
